# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

import copy
import logging
import os
import shutil
import sys
from typing import NamedTuple
from unittest import main
from unittest.mock import MagicMock

from pytest import fixture, mark, raises
from pytest_mock import MockerFixture
from test_template import TestTemplate

from mlcvzoo_util.image_io_utils import VideoLiveOutput
from mlcvzoo_util.pre_annotation_tool.pre_annotation_tool import (
    main as pre_annotation_tool_main,
)

logger = logging.getLogger(__name__)


class ProcessResult(NamedTuple):
    returncode: int


@fixture(scope="function")
def cv2_named_window(mocker: MockerFixture) -> MagicMock:
    return mocker.patch(
        "cv2.namedWindow",
        return_value=None,
    )


@fixture(scope="function")
def cv2_resize_window(mocker: MockerFixture) -> MagicMock:
    return mocker.patch(
        "cv2.resizeWindow",
        return_value=None,
    )


@fixture(scope="function")
def cv2_imshow(mocker: MockerFixture) -> MagicMock:
    return mocker.patch(
        "cv2.imshow",
        return_value=None,
    )


@fixture(scope="function")
def video_live_output_mock(mocker: MockerFixture) -> str:
    mocker.patch(
        "cv2.resizeWindow",
        return_value=None,
    )
    mocker.patch(
        "cv2.namedWindow",
        return_value=None,
    )
    mocker.patch(
        "cv2.imshow",
        return_value=None,
    )

    return mocker.patch(
        target="mlcvzoo_util.pre_annotation_tool.pre_annotation_tool.VideoLiveOutput",
        return_value=VideoLiveOutput(mode=VideoLiveOutput.MODE_GO),
    )


@fixture(scope="function")
def subprocess_run_mock(mocker: MockerFixture) -> MagicMock:
    return mocker.patch(
        "mlcvzoo_util.cvat_annotation_handler.utils.subprocess.run",
        return_value=ProcessResult(0),
    )


@fixture(scope="function")
def set_inference_mock(mocker: MockerFixture) -> MagicMock:
    return mocker.patch(
        "mlcvzoo_base.configuration.model_config.ModelConfig.set_inference",
        return_value=None,
    )


class TestPreAnnotationTool(TestTemplate):
    def __copy_test_data(self):
        original_zip_path = os.path.join(
            self.project_root,
            "test_data/test_cvat_annotation_handler/test_download/mlcvzoo-test.zip",
        )

        os.makedirs(
            os.path.join(self.project_root, "test_output/pre_annotation_tool"),
            exist_ok=True,
        )

        copy_download_zip_path = os.path.join(
            self.project_root,
            "test_output/pre_annotation_tool/mlcvzoo-test_download.zip",
        )

        copy_upload_zip_path = os.path.join(
            self.project_root,
            "test_output/pre_annotation_tool/mlcvzoo-test_upload.zip",
        )

        shutil.copy(original_zip_path, copy_download_zip_path)
        shutil.copy(original_zip_path, copy_upload_zip_path)

    @mark.usefixtures("set_inference_mock", "subprocess_run_mock")
    def test_pre_annotation_tool(self):
        self.__copy_test_data()

        argv_copy = copy.deepcopy(sys.argv)

        sys.argv = [sys.argv[0]]

        sys.argv = [
            os.path.join(
                self.project_root,
                "mlcvzoo_util/pre_annotation_tool/pre_annotation_tool.py",
            ),
            os.path.join(
                self.project_root,
                "test_data/test_pre_annotation_tool/" "test_pre_annotation_tool.yaml",
            ),
            "--replacement-config-path",
            self._gen_replacement_config(),
            "--log-dir",
            os.path.join(self.project_root, "test_output", "logs"),
            "--log-level",
            "DEBUG",
        ]

        pre_annotation_tool_main()

        sys.argv = argv_copy

    @mark.usefixtures(
        "video_live_output_mock",
        "cv2_named_window",
        "cv2_resize_window",
        "cv2_imshow",
        "subprocess_run_mock",
        "set_inference_mock",
    )
    def test_pre_annotation_tool_with_visualization(self):
        self.__copy_test_data()

        argv_copy = copy.deepcopy(sys.argv)

        sys.argv = [
            os.path.join(
                self.project_root,
                "mlcvzoo_util/pre_annotation_tool/pre_annotation_tool.py",
            ),
            os.path.join(
                self.project_root,
                "test_data/test_pre_annotation_tool",
                "test_pre_annotation_tool_with_visualization.yaml",
            ),
            "--replacement-config-path",
            self._gen_replacement_config(),
            "--log-level",
            "DEBUG",
        ]

        pre_annotation_tool_main()

        sys.argv = argv_copy

    @mark.usefixtures(
        "video_live_output_mock",
        "cv2_named_window",
        "cv2_resize_window",
        "cv2_imshow",
        "subprocess_run_mock",
        "set_inference_mock",
    )
    def test_pre_annotation_tool_wrong_model(self):
        self.__copy_test_data()

        argv_copy = copy.deepcopy(sys.argv)

        sys.argv = [
            os.path.join(
                self.project_root,
                "mlcvzoo_util/pre_annotation_tool/pre_annotation_tool.py",
            ),
            os.path.join(
                self.project_root,
                "test_data/test_pre_annotation_tool/"
                "test_pre_annotation_tool_wrong_model.yaml",
            ),
            "--replacement-config-path",
            self._gen_replacement_config(),
            "--log-level",
            "DEBUG",
        ]

        with self.assertRaises(ValueError) as value_error:
            pre_annotation_tool_main()

            assert (
                str(value_error) == "This evaluation can only be used with models "
                "that inherit from 'mlcvzoo.api.model.ObjectDetectionModel'"
            )

        sys.argv = argv_copy


if __name__ == "__main__":
    main()



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

import copy
import logging
import os
import sys
from unittest import main
from unittest.mock import MagicMock

from pytest import fixture, mark
from pytest_mock import MockerFixture
from test_template import TestTemplate

from mlcvzoo_util.model_timer.model_timer import ModelTimer
from mlcvzoo_util.model_timer.model_timer import main as model_timer_main

logger = logging.getLogger(__name__)


@fixture(scope="function")
def set_inference_mock(mocker: MockerFixture) -> MagicMock:
    return mocker.patch(
        "mlcvzoo_base.configuration.model_config.ModelConfig.set_inference",
        return_value=None,
    )


class TestModelTimer(TestTemplate):
    @mark.usefixtures(
        "set_inference_mock",
    )
    def test_on_read_from_file_object_detection_with_mlflow_logging(self) -> None:
        yaml_config_path_with_mlflow = os.path.join(
            self.project_root,
            "test_data",
            "test_model_timer",
            "model_timer-read-from-file-with-mlfow-logging.yaml",
        )

        model_timer = ModelTimer(
            configuration=ModelTimer.create_configuration(
                yaml_config_path=yaml_config_path_with_mlflow,
                string_replacement_map=self.string_replacement_map,
            )
        )

        model_timer.run()

    @mark.usefixtures(
        "set_inference_mock",
    )
    def test_on_read_from_file_object_detection_without_mlflow_logging(self) -> None:
        yaml_config_path_without_mlflow = os.path.join(
            self.project_root,
            "test_data",
            "test_model_timer",
            "model_timer-read-from-file-without-mlfow-logging.yaml",
        )

        model_timer = ModelTimer(
            configuration=ModelTimer.create_configuration(
                yaml_config_path=yaml_config_path_without_mlflow,
                string_replacement_map=self.string_replacement_map,
            )
        )

        model_timer.run()

    @mark.usefixtures(
        "set_inference_mock",
    )
    def test_model_timer_main(self):
        argv_copy = copy.deepcopy(sys.argv)

        sys.argv = [sys.argv[0]]

        sys.argv.extend(
            [
                os.path.join(
                    self.project_root,
                    "test_data/test_model_timer/"
                    "model_timer-read-from-file-with-mlfow-logging.yaml",
                ),
                "--replacement-config-path",
                self._gen_replacement_config(),
                "--log-dir",
                os.path.join(self.project_root, "test_output", "logs"),
                "--log-level",
                "DEBUG",
            ]
        )

        model_timer_main()

        sys.argv = argv_copy


if __name__ == "__main__":
    main()



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

import logging
import os
from pathlib import Path
from unittest import TestCase, main

from config_builder.replacement_map import (
    get_current_replacement_map,
    set_replacement_map_value,
    update_replacement_map_from_os,
)
from mlcvzoo_base.configuration.replacement_config import ReplacementConfig

import mlcvzoo_util

logger = logging.getLogger(__name__)


class TestTemplate(TestCase):
    def setUp(self) -> None:
        self.this_dir = Path(os.path.dirname(os.path.abspath(__file__))).resolve()

        setup_path = self.this_dir
        while setup_path.exists() and not setup_path.name == mlcvzoo_util.__name__:
            if setup_path == setup_path.parent:
                raise RuntimeError("Could not find setup_path!")
            else:
                setup_path = setup_path.parent
        # One more to be above the target directory
        setup_path = setup_path.parent

        self.project_root = str(setup_path)
        self.code_root = str(setup_path)

        set_replacement_map_value(
            ReplacementConfig.PROJECT_ROOT_DIR_KEY, self.project_root
        )
        update_replacement_map_from_os()
        self.string_replacement_map = get_current_replacement_map()

        logger.debug(
            "Setup finished: \n"
            " - this_dir: %s\n"
            " - project_root: %s\n"
            " - code_root: %s\n"
            % (
                self.this_dir,
                self.project_root,
                self.code_root,
            )
        )
        os.chdir(self.code_root)
        logger.info("Changed working directory to: '%s'", self.code_root)

    def _gen_replacement_config(self) -> str:
        path = Path("%s/test_output/" % self.project_root)
        # project_root should point to an already existing location, so do not create parents here
        path.mkdir(exist_ok=True, parents=False)
        path = path / "replacement_config.yaml"
        with open(path, "w", encoding="utf-8") as file:
            file.write("PROJECT_ROOT_DIR: %s\n" % self.project_root)
            return str(path)


if __name__ == "__main__":
    main()



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Module for parsing information from yaml in python accessible attributes for the ModelSuite class.
"""

import related
from attr import define
from config_builder import BaseConfigClass
from mlcvzoo_base.configuration.model_config import ModelConfig


@define
class ModelTrainerConfig(BaseConfigClass):
    """
    Class for parsing general information about the model suite and also further information
    in respective hierarchy
    """

    __related_strict__ = True

    model_config: ModelConfig = related.ChildField(cls=ModelConfig)



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Central module for handling the download and upload
of zip files to CVAT via their commandline interface
"""

import argparse
import logging
import sys
from typing import Dict, Optional, cast

from config_builder import ConfigBuilder

from mlcvzoo_util.cvat_annotation_handler.configuration import (
    CVATAnnotationHandlerConfig,
)
from mlcvzoo_util.cvat_annotation_handler.cvat_dumper import CVATDumper
from mlcvzoo_util.cvat_annotation_handler.cvat_uploader import PascalVOCUploader
from mlcvzoo_util.logger import Logger
from mlcvzoo_util.mlcvzoo_cli_tool import MLCVZooCLITool

logger = logging.getLogger(__name__)


class CVATAnnotationHandler(MLCVZooCLITool[CVATAnnotationHandlerConfig]):
    """
    Central class for handling the download and upload
    of zip files to CVAT via their commandline interface
    """

    def __init__(self, configuration: CVATAnnotationHandlerConfig):
        """
        Instantiates a CVATAnnotationHandler object

        Args:
            configuration: (Optional) An already existing configuration object
        """

        self.configuration: CVATAnnotationHandlerConfig = configuration

    @staticmethod
    def create_configuration(
        yaml_config_path: str,
        string_replacement_map: Optional[Dict[str, str]] = None,
        no_checks: bool = False,
    ) -> CVATAnnotationHandlerConfig:
        """
        Create a CVATAnnotationHandlerConfig

        Args:
            yaml_config_path:  A yaml filepath where to build the configuration
                               object from
            string_replacement_map: A dictionary that defines placeholders which can be used
                                    while parsing the file. They can be understood as variables
                                    that can be used to define configs that are valid across
                                    multiple devices.
            no_checks: Whether the configuration object should be checked for mutual exclusiveness
                       and the "check_values" method for each attribute of the supertype
                       "BaseConfigClass" should be called
        Returns:
            The created configuration
        """

        return cast(
            CVATAnnotationHandlerConfig,
            ConfigBuilder(
                class_type=CVATAnnotationHandlerConfig,
                yaml_config_path=yaml_config_path,
                string_replacement_map=string_replacement_map,
                no_checks=no_checks,
            ).configuration,
        )

    @staticmethod
    def create_cli_configuration(
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> CVATAnnotationHandlerConfig:
        """
        Create a CVATAnnotationHandlerConfig

        Args:
            string_replacement_map: A dictionary that defines placeholders which can be used
                                    while parsing the file. They can be understood as variables
                                    that can be used to define configs that are valid across
                                    multiple devices.
        Returns:
            The created configuration
        """

        return MLCVZooCLITool._create_cli_configuration(
            configuration_class=CVATAnnotationHandlerConfig,
            string_replacement_map=string_replacement_map,
            configure_argparse=CVATAnnotationHandler.configure_argparse,
        )

    def download_all_tasks(self) -> None:
        """
        Execute all downloads that are specified in the configuration object
        of this class

        Returns:
            None
        """

        for dump_task_config in self.configuration.dump_task_configs:
            CVATDumper.dump_task_data(
                dump_task_config=dump_task_config,
                cvat_cli_config=self.configuration.cvat_cli_config,
                disable_ssl_verify=self.configuration.cvat_cli_config.disable_ssl_verify,
            )
            logger.info("==========  DUMP TASK FINISHED  ==========\n")

    def upload_all_tasks(self) -> None:
        """
        Execute all uploads that are specified in the configuration object
        of this class

        Returns:
            None
        """

        for upload_task_config in self.configuration.upload_task_configs:
            PascalVOCUploader.upload_task_data(
                upload_task_config=upload_task_config,
                cvat_cli_config=self.configuration.cvat_cli_config,
                disable_ssl_verify=self.configuration.cvat_cli_config.disable_ssl_verify,
            )
            logger.info("==========  UPLOAD TASK FINISHED  ==========\n")

    def run(self) -> None:
        """
        Execute all downloads and uploads that are specified in the configuration object
        of this class

        Returns:
            None
        """

        logger.info("Start to download and upload tasks")

        self.download_all_tasks()
        self.upload_all_tasks()

        logger.info("Finished to download and upload tasks")

    @staticmethod
    def configure_argparse(
        parser: argparse.ArgumentParser,
    ) -> None:
        MLCVZooCLITool.configure_argparse(parser)
        parser.description = "Download and upload annotation files from/to CVAT"


def main() -> None:
    """
    Main entry point of the CVATAnnotationHandler tool

    Returns:
        None
    """

    args = ConfigBuilder.setup_argparse(
        configure_argparse=CVATAnnotationHandler.configure_argparse
    ).parse_args()

    Logger.init_logging_basic(
        log_dir=args.log_dir,
        log_file_postfix="CVATAnnotationHandler",
        no_stdout=False,
        root_log_level=args.log_level,
    )

    cvat_annotation_handler = CVATAnnotationHandler(
        configuration=CVATAnnotationHandler.create_cli_configuration()
    )
    cvat_annotation_handler.run()


if __name__ == "__main__":
    main()



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module for evaluating mlcvzoo models"""

import argparse
import logging
import os.path
from abc import ABC
from typing import Any, Dict, Generic, List, Optional, Tuple, Type, cast

import mlflow
from config_builder import ConfigBuilder
from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.interfaces import Classifiable, NetBased
from mlcvzoo_base.api.model import Model
from mlcvzoo_base.configuration.annotation_handler_config import AnnotationHandlerConfig
from mlcvzoo_base.configuration.mlfow_config import MLFlowConfig
from mlcvzoo_base.configuration.model_config import ModelConfig
from mlcvzoo_base.configuration.structs import MLFlowExperimentTypes
from mlcvzoo_base.data_preparation.annotation_handler import AnnotationHandler
from mlcvzoo_base.metrics.mlflow.mlflow_runner import MLFLowRunner
from mlcvzoo_base.models.model_registry import ModelRegistry
from mlcvzoo_base.utils import get_file_list

from mlcvzoo_util.logger import Logger
from mlcvzoo_util.mlcvzoo_cli_tool import MLCVZooCLITool, configure_model_argparse
from mlcvzoo_util.model_evaluator.configuration import (
    CheckpointConfig,
    ModelEvaluatorCLIConfig,
    ModelEvaluatorConfig,
)
from mlcvzoo_util.model_evaluator.metric_factory import (
    EvaluationMetricType,
    MetricFactory,
    ModelType,
    get_factory,
)
from mlcvzoo_util.model_evaluator.structs import CheckpointInfo, CheckpointLoggingModes

logger = logging.getLogger(__name__)


class ModelEvaluator(
    MLCVZooCLITool[ModelEvaluatorConfig], ABC, Generic[EvaluationMetricType, ModelType]
):
    """
    Tool for evaluating NetBased MLCVZoo models. It gathers all checkpoints that have
    been generated by training run and determines the checkpoint with regard to the type
    of the respective Model.
    There are two kind of usages:
    1. Use it programmatically
    2. Use it as a commandline tool

    Additionally, the training parameters, the calculated metrics and selected checkpoints
    can be logged with mlflow.
    """

    @staticmethod
    def create_configuration(
        yaml_config_path: str,
        string_replacement_map: Optional[Dict[str, str]] = None,
        no_checks: bool = False,
    ) -> ModelEvaluatorConfig:
        """
        Create a ModelEvaluatorConfig

        Args:
            yaml_config_path:  A yaml filepath where to build the configuration
                               object from
            string_replacement_map: A dictionary that defines placeholders which can be used
                                    while parsing the file. They can be understood as variables
                                    that can be used to define configs that are valid across
                                    multiple devices.
            no_checks: Whether the configuration object should be checked for mutual exclusiveness
                       and the "check_values" method for each attribute of the supertype
                       "BaseConfigClass" should be called
        Returns:
            The created configuration
        """

        return cast(
            ModelEvaluatorConfig,
            ConfigBuilder(
                class_type=ModelEvaluatorConfig,
                yaml_config_path=yaml_config_path,
                string_replacement_map=string_replacement_map,
                no_checks=no_checks,
            ).configuration,
        )

    @staticmethod
    def create_cli_configuration(
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> ModelEvaluatorCLIConfig:
        """
        Create a ModelEvaluatorConfig

        Args:
            string_replacement_map: A dictionary that defines placeholders which can be used
                                    while parsing the file. They can be understood as variables
                                    that can be used to define configs that are valid across
                                    multiple devices.
        Returns:
            The created configuration
        """

        return MLCVZooCLITool._create_cli_configuration(
            configuration_class=ModelEvaluatorCLIConfig,
            string_replacement_map=string_replacement_map,
            configure_argparse=ModelEvaluator.configure_argparse,
        )

    @staticmethod
    def configure_argparse(
        parser: argparse.ArgumentParser,
    ) -> None:
        parser.description = (
            "Runs an evaluation on the configured models "
            "using the configuration evaluation data"
        )
        MLCVZooCLITool.configure_argparse(parser)
        configure_model_argparse(parser=parser)
        parser.add_argument(
            "--single-mode",
            default=False,
            action="store_true",
            help="Whether to use the checkpoint-config of the ModelEvaluator configuration "
            "or the configured inference checkpoint of the model for determining the"
            "checkpoints that should be evaluated.",
        )

    @staticmethod
    def create_model(
        model_config: ModelConfig,
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> ModelType:
        """
        Creates a model based on the given config

        Args:
            model_config: The model config that should be used
                          for creating a model
            string_replacement_map: (Optional) A dictionary that defines placeholders which can
                                    be used while parsing a configuration file. They can be
                                    understood as variables that can be used to define configs
                                    that are valid across multiple devices.
        Returns:
            The created model
        """
        # In order to be able to load the checkpoints that
        # should be evaluated, we need an inference model
        model_config.set_inference(inference=True)
        return ModelRegistry().init_model(  # type: ignore[return-value]
            model_config=model_config, string_replacement_map=string_replacement_map
        )

    @staticmethod
    def generate_evaluation_annotations(
        annotation_handler_config: AnnotationHandlerConfig,
        mapper: Optional[AnnotationClassMapper] = None,
    ) -> List[BaseAnnotation]:
        gt_annotations: List[BaseAnnotation] = AnnotationHandler(
            mapper=mapper,
            configuration=annotation_handler_config,
        ).parse_inference_annotations()

        if mapper is not None:
            gt_annotations = AnnotationHandler.reduce_annotations(
                annotations=gt_annotations, mapper=mapper
            )

        return gt_annotations

    @staticmethod
    def select_evaluation_checkpoints(
        checkpoint_config: CheckpointConfig,
    ) -> List[str]:
        checkpoint_paths = get_file_list(
            input_dir=checkpoint_config.checkpoint_dir,
            file_extension=checkpoint_config.checkpoint_filename_suffix,
        )
        # Sort the checkpoints assuming that they will be sorted in ascending order
        # by epoch
        checkpoint_paths.sort()

        return [c for c in checkpoint_paths if c not in checkpoint_config.ignore]

    @staticmethod
    def _restore_checkpoint(checkpoint_path: str, model: Model) -> None:  # type: ignore[type-arg]
        if not os.path.exists(checkpoint_path):
            raise ValueError(
                f"The given checkpoint path '{checkpoint_path}' does not exist! "
                "Only absolute paths of checkpoints are allowed!"
            )

        if not isinstance(model, NetBased):
            raise ValueError(
                "This evaluation can only be used with models that "
                "inherit from 'mlcvzoo.api.interfaces.NetBased'"
            )

        model.restore(checkpoint_path=checkpoint_path)

    def _post_evaluation_step(
        self,
        checkpoint_path: str,
        evaluated_checkpoint_metrics: Dict[str, EvaluationMetricType],
        inference_model: ModelType,
    ) -> None:
        pass

    def evaluate_checkpoints(
        self,
        model_evaluator_config: ModelEvaluatorConfig,
        inference_model: ModelType,
        gt_annotations: List[BaseAnnotation],
        checkpoint_paths: Optional[List[str]],
        log_results: bool = False,
    ) -> Tuple[Dict[str, EvaluationMetricType], CheckpointInfo]:
        """
        Evaluate all given checkpoints according to the given configuration.
        If specified in the model-evaluator-config log selected parameters,
        checkpoints and metrics with mlflow.

        Args:
            model_evaluator_config: ModelEvaluatorConfig for this evaluation run
            inference_model: ObjectDetectionModel instantiated for inference
            gt_annotations: Ground truth annotations created with the mapper of the
                            given inference-model
            checkpoint_paths: List of checkpoints that should be evaluated
            log_results: Whether to log the results to mlflow and tensorboard

        Returns:
            Evaluated checkpoints and their calculated metrics, info about best checkpoint
        """

        metric_factory = get_factory(inference_model=inference_model)

        if metric_factory is None:
            raise ValueError(
                "Could not determine a metric-factory for "
                f"model-type='{type(inference_model)}'"
            )

        evaluated_checkpoint_metrics: Dict[str, EvaluationMetricType] = {}
        if checkpoint_paths:
            for ckpt in checkpoint_paths:
                ModelEvaluator._restore_checkpoint(
                    checkpoint_path=ckpt, model=inference_model
                )

                evaluated_checkpoint_metrics[ckpt] = metric_factory.compute_metrics(
                    inference_model=inference_model,
                    gt_annotations=gt_annotations,
                    model_evaluator_config=model_evaluator_config,
                )

                self._post_evaluation_step(
                    checkpoint_path=ckpt,
                    evaluated_checkpoint_metrics=evaluated_checkpoint_metrics,
                    inference_model=inference_model,
                )
        else:
            # TODO: Use interface method for getting the checkpoint-path of the model
            evaluated_checkpoint_metrics[MetricFactory.MODEL_STATE_INDICATOR] = (
                metric_factory.compute_metrics(
                    inference_model=inference_model,
                    gt_annotations=gt_annotations,
                    model_evaluator_config=model_evaluator_config,
                )
            )

            self._post_evaluation_step(
                checkpoint_path=MetricFactory.MODEL_STATE_INDICATOR,
                evaluated_checkpoint_metrics=evaluated_checkpoint_metrics,
                inference_model=inference_model,
            )

        best_checkpoint_info = metric_factory.determine_best_checkpoint(
            evaluated_checkpoint_metrics=evaluated_checkpoint_metrics,
        )

        logger.info(
            "Evaluation finished, best checkpoint:\n" "\t- ckpt='%s'\n" "\t- ap='%s'\n",
            best_checkpoint_info.path,
            best_checkpoint_info.score,
        )

        if log_results:
            if model_evaluator_config.mlflow_config is not None:
                ModelEvaluator._start_mlflow(
                    mlflow_config=model_evaluator_config.mlflow_config.config,
                    model_name=inference_model.unique_name,
                )
                checkpoint_log_mode: str = (
                    model_evaluator_config.mlflow_config.checkpoint_log_mode
                )
            else:
                checkpoint_log_mode = CheckpointLoggingModes.BEST.value

            metric_factory.log_results(
                checkpoint_log_mode=checkpoint_log_mode,
                evaluated_checkpoint_metrics=evaluated_checkpoint_metrics,
                best_checkpoint=best_checkpoint_info,
                logging_configs=(
                    [model_evaluator_config.tensorboard_logging_config]
                    if model_evaluator_config.tensorboard_logging_config
                    else []
                ),
            )
            if model_evaluator_config.mlflow_config is not None:
                mlflow.end_run()

        return evaluated_checkpoint_metrics, best_checkpoint_info

    @staticmethod
    def _start_mlflow(mlflow_config: MLFlowConfig, model_name: str) -> None:
        runner = MLFLowRunner(configuration=mlflow_config)
        runner.create_mlflow_experiments()

        runner.start_mlflow_run(
            experiment_name=MLFlowExperimentTypes.EVAL,
            run_name=model_name,
        )


def __get_model_checkpoint(model: Model) -> Optional[List[str]]:  # type: ignore
    if isinstance(model, NetBased) and isinstance(model, Model):
        try:
            # TODO: Realize this as an API method?
            return [model.configuration.inference_config.checkpoint_path]
        except AttributeError:
            pass
    else:
        logger.warning(
            "Could not determine the model-checkpoint, "
            "this feature is only available for models that "
            "inherit from 'mlcvzoo.api.model.Model' and 'mlcvzoo.api.interfaces.NetBased"
        )

    return None


def run_evaluation(
    configuration: ModelEvaluatorCLIConfig,
    single_mode: bool,
) -> Tuple[Dict[str, EvaluationMetricType], CheckpointInfo]:
    """
    Runs an generic mlcvzoo evaluation

    Args:
        configuration:
        single_mode: Whether
        model:

    Returns:
        None
    """
    model: Model[Any, Any, Any] = ModelEvaluator.create_model(
        model_config=configuration.model_config,
        string_replacement_map=configuration.string_replacement_map,
    )

    if not isinstance(model, Classifiable) or not isinstance(model, Model):
        raise ValueError(
            "This evaluation can only be used with models that "
            "inherit from 'mlcvzoo.api.model.Model' and 'mlcvzoo.api.interfaces.Classifiable"
        )

    assert isinstance(model, Classifiable) and isinstance(model, Model)
    gt_annotations: List[BaseAnnotation] = (
        ModelEvaluator.generate_evaluation_annotations(
            annotation_handler_config=configuration.annotation_handler_config,
            mapper=model.mapper,
        )
    )

    checkpoint_paths: Optional[List[str]]

    if single_mode:
        checkpoint_paths = __get_model_checkpoint(model=model)
    else:
        checkpoint_paths = ModelEvaluator.select_evaluation_checkpoints(
            checkpoint_config=configuration.checkpoint_config
        )

    if checkpoint_paths is None:
        logger.warning(
            "Could not determine any model checkpoints, therefore model"
            "will be used as is for the evaluation"
        )

    return ModelEvaluator().evaluate_checkpoints(
        model_evaluator_config=configuration,
        inference_model=model,  # type: ignore[arg-type]
        gt_annotations=gt_annotations,
        checkpoint_paths=checkpoint_paths,
        log_results=True,
    )


def main() -> None:
    """
    Entry point when using the mlcvzoo-modelevaluator command line tool.
    (See [tool.poetry.scripts] section in pyproject.toml)

    Returns:
        None
    """

    args = ConfigBuilder.setup_argparse(
        configure_argparse=ModelEvaluator.configure_argparse
    ).parse_args()

    Logger.init_logging_basic(
        log_dir=args.log_dir,
        log_file_postfix="ModelEvaluator",
        no_stdout=False,
        root_log_level=args.log_level,
    )

    configuration = ModelEvaluator.create_cli_configuration()

    configuration.model_config.update_class_type(
        args_dict=vars(args),
    )

    model_type: Optional[Type[Model[Any, Any, Any]]] = ModelRegistry().get_model_type(
        class_type=configuration.model_config.class_type
    )

    if model_type:
        configuration.model_config.update_constructor_parameters(
            args_dict=vars(args),
            model_type=model_type,
        )
    else:
        raise ValueError(
            f"Model type '{configuration.model_config.class_type}' is not registered. "
            f"Use one of '{list(ModelRegistry().get_registered_models().keys())}'",
        )

    run_evaluation(configuration=configuration, single_mode=args.single_mode)


if __name__ == "__main__":
    main()



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module for handling the download of zip files to CVAT via their commandline interface"""

import logging
import os
import shutil

import yaml
from mlcvzoo_base.utils import ensure_dir, get_file_list
from mlcvzoo_base.utils.file_utils import extract_zip_data
from related import to_yaml

from mlcvzoo_util.cvat_annotation_handler.configuration import (
    CVATCLIConfig,
    CVATTaskDumpConfig,
    CVATTaskInfoConfig,
)
from mlcvzoo_util.cvat_annotation_handler.utils import run_command_in_process

logger = logging.getLogger(__name__)


class CVATDumper:
    """
    Class for handling the download of zip files to CVAT via their commandline interface
    """

    @staticmethod
    def dump_task_data(
        dump_task_config: CVATTaskDumpConfig,
        cvat_cli_config: CVATCLIConfig,
        disable_ssl_verify: bool = False,
    ) -> None:
        """
        Execute the download of a single zip file to the CVAT instance

        Args:
            dump_task_config: Configuration of the task that is to be downloaded
            cvat_cli_config: Basis configuration that is needed for the communication between
                             the commandline-interface and CVAT
            disable_ssl_verify: Whether to allow uncertified https connections

        Returns:
            None
        """

        logger.info(
            "\n==========  DUMP TASK  ==========\n" "'''\n%s'''\n",
            to_yaml(dump_task_config, yaml_package=yaml, dumper_cls=yaml.Dumper),
        )

        target_zip_path = CVATDumper.__create_target_zip_path(
            dump_task_config=dump_task_config
        )

        if CVATDumper.__pre_clean_up_and_determine_skip(
            overwrite_existing_zip=dump_task_config.overwrite_existing_zip,
            target_zip_path=target_zip_path,
        ):
            return

        # ================================
        # DOWNLOAD TASK DATA
        CVATDumper.call_cvat_dump_via_cli(
            cvat_cli_config=cvat_cli_config,
            target_zip_path=target_zip_path,
            task_info=dump_task_config.task_info,
            disable_ssl_verify=disable_ssl_verify,
        )

        zip_extract_dir = CVATDumper.__create_zip_extract_dir_path(
            target_zip_path=target_zip_path,
            zip_extract_dir=dump_task_config.zip_extract_dir,
        )

        # ================================
        # extract and move data
        if dump_task_config.extract_data:
            extract_zip_data(
                zip_path=target_zip_path,
                zip_extract_dir=zip_extract_dir,
            )

        if dump_task_config.move_extracted_data_to_dataset_dir is not None:
            CVATDumper.__extract_and_copy_annotations_to_data_dir(
                target_zip_path=target_zip_path,
                zip_extract_dir=zip_extract_dir,
                dataset_dir=dump_task_config.dataset_dir,
            )

        # ================================
        # post clean up
        if dump_task_config.clean_up_dump_data:
            CVATDumper.__post_clean_up(
                target_zip_path=target_zip_path,
                zip_extract_dir=zip_extract_dir,
            )

    @staticmethod
    def call_cvat_dump_via_cli(
        cvat_cli_config: CVATCLIConfig,
        target_zip_path: str,
        task_info: CVATTaskInfoConfig,
        disable_ssl_verify: bool = False,
    ) -> None:
        """
        Download a zip file from CVAT.

        Args:
            cvat_cli_config: Basis configuration that is needed for the communication between
                             the commandline-interface and CVAT
            target_zip_path: The zip file to upload
            task_info: Configuration for the specific task
            disable_ssl_verify: Whether to allow uncertified https connections

        Returns:
            None
        """

        ensure_dir(file_path=target_zip_path, verbose=True)
        command = CVATDumper.__create_cvat_dump_cli_command(
            base_command=cvat_cli_config.create_base_command_string(),
            target_zip_path=target_zip_path,
            task_info=task_info,
        )

        logger.info(
            "Dump annotation data: Task-ID=%s, " "zip-path=%s",
            task_info.task_ID,
            target_zip_path,
        )
        return_code = run_command_in_process(
            command=command, disable_ssl_verify=disable_ssl_verify
        )

        if return_code != 0:
            raise IOError(f"CVAT download did not succeed, return-code={return_code}")

    @staticmethod
    def __create_target_zip_path(dump_task_config: CVATTaskDumpConfig) -> str:
        if dump_task_config.target_zip_path == "":
            target_zip_path: str = os.path.join("tmp", "tmp-task.zip")
            logger.debug(
                "dump_task_config.target_zip_path is not set, therefore the default '%s' "
                "will be used to store the downloaded task:",
                target_zip_path,
            )
        else:
            target_zip_path = dump_task_config.target_zip_path

        return target_zip_path

    @staticmethod
    def __create_zip_extract_dir_path(
        target_zip_path: str, zip_extract_dir: str
    ) -> str:
        if not os.path.isdir(zip_extract_dir):
            zip_extract_dir = target_zip_path.replace(".zip", "")
            logger.debug(
                "dump_task_config.target_zip_path is not set, therefore the default '%s' "
                "will be used to store the downloaded task:",
                zip_extract_dir,
            )

        return zip_extract_dir

    @staticmethod
    def __pre_clean_up_and_determine_skip(
        overwrite_existing_zip: bool, target_zip_path: str
    ) -> bool:
        if os.path.isfile(target_zip_path):
            if overwrite_existing_zip:
                logger.info("Clean up old dump data, remove '%s'", target_zip_path)
                os.remove(target_zip_path)
            else:
                logger.info(
                    "Skip download since zip-file '%s' exists and should not be overwritten",
                    target_zip_path,
                )
                return True

        return False

    @staticmethod
    def __post_clean_up(target_zip_path: str, zip_extract_dir: str) -> None:
        if os.path.isfile(target_zip_path):
            logger.debug("(DUMP) Clean up zip file: '%s'", target_zip_path)
            os.remove(target_zip_path)

        if os.path.isdir(zip_extract_dir):
            logger.debug(
                "(DUMP) Clean up zip extraction directory: '%s'", zip_extract_dir
            )
            shutil.rmtree(zip_extract_dir)

    @staticmethod
    def __create_cvat_dump_cli_command(
        base_command: str, target_zip_path: str, task_info: CVATTaskInfoConfig
    ) -> str:
        return (
            f"{base_command} dump "
            f"{task_info.task_ID} "
            f"{target_zip_path} "
            f"--format '{task_info.annotation_format}'"
        )

    @staticmethod
    def __extract_and_copy_annotations_to_data_dir(
        target_zip_path: str,
        zip_extract_dir: str,
        dataset_dir: str,
    ) -> None:
        if not os.path.isdir(dataset_dir):
            logger.warning(
                "Path to dataset_dir='%s' does not exist! "
                "Skip the extraction and move of annotation data.",
                dataset_dir,
            )
            return

        extract_zip_data(zip_path=target_zip_path, zip_extract_dir=zip_extract_dir)

        zip_annotation_dir = os.path.join(zip_extract_dir, "Annotations")

        annotation_file_paths = get_file_list(
            input_dir=zip_extract_dir, search_subfolders=True, file_extension=".xml"
        )

        for annotation_path in annotation_file_paths:
            dataset_annotation_path = annotation_path.replace(
                zip_annotation_dir, dataset_dir
            )

            ensure_dir(file_path=dataset_annotation_path, verbose=True)

            logger.debug(
                "Copy annotation '%s' to zip '%s'",
                annotation_path,
                dataset_annotation_path,
            )

            shutil.copy(src=annotation_path, dst=dataset_annotation_path)



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

import logging
from unittest import main
from unittest.mock import MagicMock

from pytest import fixture, mark
from pytest_mock import MockerFixture
from test_template import TestTemplate

from mlcvzoo_util.image_io_utils import VideoLiveOutput

logger = logging.getLogger(__name__)


@fixture(scope="function")
def cv2_named_window(mocker: MockerFixture) -> MagicMock:
    return mocker.patch(
        "cv2.namedWindow",
        return_value=None,
    )


class TestImageIOUtils(TestTemplate):
    @mark.usefixtures("cv2_named_window")
    def test_video_live_output_constructor(self) -> None:
        VideoLiveOutput()

    @mark.usefixtures("cv2_named_window")
    def test_video_live_output_constructor_mode(self) -> None:
        VideoLiveOutput(mode=VideoLiveOutput.MODE_GO)

    @mark.usefixtures("cv2_named_window")
    def test_video_live_output_constructor_mode_fail(self) -> None:
        with self.assertRaises(ValueError) as value_error:
            VideoLiveOutput(mode="other")

            assert (
                str(value_error) == "Invalid mode='other' "
                "has to be one of '['terminate', 'go', 'step']'"
            )


if __name__ == "__main__":
    main()



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module for handling the training of mlcvzoo models"""

import argparse
import logging
from typing import Any, Dict, Optional, Type, cast

from config_builder import ConfigBuilder
from mlcvzoo_base.api.interfaces import Trainable
from mlcvzoo_base.api.model import Model
from mlcvzoo_base.configuration.model_config import ModelConfig
from mlcvzoo_base.models.model_registry import ModelRegistry

from mlcvzoo_util.logger import Logger
from mlcvzoo_util.mlcvzoo_cli_tool import MLCVZooCLITool, configure_model_argparse
from mlcvzoo_util.model_trainer.configuration import ModelTrainerConfig

logger = logging.getLogger(__name__)


class ModelTrainer(MLCVZooCLITool[ModelTrainerConfig]):
    """Class to handle the training of different models"""

    def __init__(
        self,
        configuration: ModelTrainerConfig,
    ):
        self.configuration: ModelTrainerConfig = configuration

    @staticmethod
    def create_configuration(
        yaml_config_path: str,
        string_replacement_map: Optional[Dict[str, str]] = None,
        no_checks: bool = False,
    ) -> ModelTrainerConfig:
        """
        Create a ModelTrainerConfig

        Args:
            yaml_config_path:  A yaml filepath where to build the configuration
                               object from
            string_replacement_map: A dictionary that defines placeholders which can be used
                                    while parsing the file. They can be understood as variables
                                    that can be used to define configs that are valid across
                                    multiple devices.
            no_checks: Whether the configuration object should be checked for mutual exclusiveness
                       and the "check_values" method for each attribute of the supertype
                       "BaseConfigClass" should be called
        Returns:
            The created configuration
        """

        return cast(
            ModelTrainerConfig,
            ConfigBuilder(
                class_type=ModelTrainerConfig,
                yaml_config_path=yaml_config_path,
                string_replacement_map=string_replacement_map,
                no_checks=no_checks,
            ).configuration,
        )

    @staticmethod
    def create_cli_configuration(
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> ModelTrainerConfig:
        """
        Create a ModelTrainerConfig

        Args:
            string_replacement_map: A dictionary that defines placeholders which can be used
                                    while parsing the file. They can be understood as variables
                                    that can be used to define configs that are valid across
                                    multiple devices.
        Returns:
            The created configuration
        """

        return MLCVZooCLITool._create_cli_configuration(
            configuration_class=ModelTrainerConfig,
            string_replacement_map=string_replacement_map,
            configure_argparse=ModelTrainer.configure_argparse,
        )

    @staticmethod
    def configure_argparse(
        parser: argparse.ArgumentParser,
    ) -> None:
        parser.description = "Run training for the given model-config"
        MLCVZooCLITool.configure_argparse(parser)
        configure_model_argparse(parser=parser)

    @staticmethod
    def create_model(
        model_config: ModelConfig,
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> Model:  # type: ignore[type-arg]
        """
        Creates a model based on the given config

        Args:
            model_config: The model config that should be used
                          for creating a model
            string_replacement_map: (Optional) A dictionary that defines placeholders which can
                                    be used while parsing a configuration file. They can be
                                    understood as variables that can be used to define configs
                                    that are valid across multiple devices.
        Returns:
            The created model
        """
        # In order to be able to load the checkpoints that
        # should be evaluated, we need an inference model
        model_config.set_inference(inference=False)
        return ModelRegistry().init_model(
            model_config=model_config, string_replacement_map=string_replacement_map
        )

    def run_training(self) -> Model[Any, Any, Any]:
        """
        Runs the training with each of the models in the models attribute

        Returns:
            The model object that has been trained
        """

        model = ModelTrainer.create_model(
            model_config=self.configuration.model_config,
            string_replacement_map=self.configuration.string_replacement_map,
        )

        if not isinstance(model, Trainable):
            raise ValueError(
                "The ModelTrainer only works with models that inherit from the"
                "Trainable interface"
            )
        model.train()

        return model


def main() -> None:
    """
    Entry point when using the mlcvzoo-modeltrainer command line tool.
    (See [tool.poetry.scripts] section in pyproject.toml)
    """

    args = ConfigBuilder.setup_argparse(
        configure_argparse=ModelTrainer.configure_argparse
    ).parse_args()

    Logger.init_logging_basic(
        log_dir=args.log_dir,
        log_file_postfix="ModelTimer",
        no_stdout=False,
        root_log_level=args.log_level,
    )

    configuration: ModelTrainerConfig = ModelTrainer.create_cli_configuration()

    configuration.model_config.update_class_type(
        args_dict=vars(args),
    )

    model_type: Optional[Type[Model[Any, Any, Any]]] = ModelRegistry().get_model_type(
        class_type=configuration.model_config.class_type
    )

    if model_type:
        configuration.model_config.update_constructor_parameters(
            args_dict=vars(args),
            model_type=model_type,
        )
    else:
        raise ValueError(
            f"Model type '{configuration.model_config.class_type}' is not registered. "
            f"Use one of '{list(ModelRegistry().get_registered_models().keys())}'",
        )

    model_timer = ModelTrainer(configuration=configuration)
    model_timer.run_training()


if __name__ == "__main__":
    main()



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

import copy
import logging
import os
import shlex
import shutil
import subprocess
import sys
import xml.etree.ElementTree as ET_xml
from pathlib import Path
from typing import NamedTuple
from unittest import main
from unittest.mock import MagicMock

import pytest
from mlcvzoo_base.utils.file_utils import extract_zip_data
from pytest import fixture
from pytest_mock import MockerFixture
from test_template import TestTemplate

from mlcvzoo_util.cvat_annotation_handler.configuration import (
    CVATCLIConfig,
    CVATTaskDumpConfig,
    CVATTaskInfoConfig,
    CVATTaskUploadConfig,
)
from mlcvzoo_util.cvat_annotation_handler.cvat_annotation_handler import (
    CVATAnnotationHandler,
)
from mlcvzoo_util.cvat_annotation_handler.cvat_annotation_handler import (
    main as cvat_annotation_handler_main,
)
from mlcvzoo_util.cvat_annotation_handler.cvat_dumper import CVATDumper
from mlcvzoo_util.cvat_annotation_handler.cvat_uploader import PascalVOCUploader

logger = logging.getLogger(__name__)


def _xml_equal(xml_path_1: str, xml_path_2: str) -> bool:
    """
    Helper for comparison of two XML trees

    Args:
        xml_path_1: path to one xml file
        xml_path_2: path to the other xml file

    Returns:
        True if xml trees are identical, else False.
    """

    tree_1 = ET_xml.parse(xml_path_1)
    root_1 = tree_1.getroot()

    tree_2 = ET_xml.parse(xml_path_2)
    root_2 = tree_2.getroot()

    return _xml_root_compare(root_1=root_1, root_2=root_2)


def _xml_root_compare(root_1: ET_xml.Element, root_2: ET_xml.Element) -> bool:
    """
    Recursive helper function that compares tags and attributes of a tree and
     calls itself for each child.
    Args:
        root_1: xml.etree.ElementTree.Element object
        root_2: xml.etree.ElementTree.Element object

    Returns:
        True if trees are identical, else False.

    """

    if root_1.tag == root_2.tag:
        # NOTE: The local and gitlab-ci "path" tag are different, but are
        #       not relevant for this test
        if root_1.tag == "path":
            result = True
        elif root_1.tag == root_2.tag and root_1.text == root_2.text:
            result = True

            for index, child_1 in enumerate(root_1):
                if len(root_2) > index:
                    child_2 = root_2[index]
                    result = result and _xml_root_compare(child_1, child_2)
                else:
                    # If both roots do have the same length, they are not equal
                    result = False
                    break
        else:
            result = False
    else:
        result = False

    return result


class ProcessResult(NamedTuple):
    returncode: int


@fixture(scope="function")
def subprocess_run_mock(mocker: MockerFixture) -> MagicMock:
    return mocker.patch(
        "mlcvzoo_util.cvat_annotation_handler.utils.subprocess.run",
        return_value=ProcessResult(0),
    )


@fixture(scope="function")
def subprocess_run_failed_1_mock(mocker: MockerFixture) -> MagicMock:
    return mocker.patch(
        "mlcvzoo_util.cvat_annotation_handler.utils.subprocess.run",
        return_value=ProcessResult(1),
    )


@fixture(scope="function")
def os_path_isdir_mock(mocker: MockerFixture) -> MagicMock:
    return mocker.patch(
        "os.path.isdir",
        return_value=True,
    )


@fixture(scope="function")
def os_path_isdir_not_mock(mocker: MockerFixture) -> MagicMock:
    return mocker.patch(
        "os.path.isdir",
        return_value=False,
    )


@fixture(scope="function")
def shutil_rmtree_mock(mocker: MockerFixture) -> MagicMock:
    return mocker.patch(
        "shutil.rmtree",
    )


@fixture(scope="function")
def getpass_mock(mocker: MockerFixture) -> MagicMock:
    return mocker.patch(
        "mlcvzoo_util.cvat_annotation_handler.configuration.getpass",
        return_value="PASSWORD",
    )


class TestCVATAnnotationHandler(TestTemplate):
    def setUp(self) -> None:
        TestTemplate.setUp(self)
        self.string_replacement_map["HOME"] = str(Path.home())

    def tearDown(self) -> None:
        shutil.rmtree(
            path=os.path.join(
                self.project_root, "test_output/test_cvat_annotation_handler"
            ),
            ignore_errors=True,
        )

    def test_cvat_annotation_handler_correct_base_command(self):
        cvat_cli_config = CVATCLIConfig(
            cli_path="CVAT_DIR/utils/cli/cli.py",
            server_host="http://localhost",
            server_port=8080,
            disable_ssl_verify=True,
            auth="USERNAME:PASSWORD",
        )

        expected_base_command = (
            f"{sys.executable} "
            "CVAT_DIR/utils/cli/cli.py "
            "--server-host http://localhost "
            "--server-port 8080 "
            "--auth USERNAME:PASSWORD"
        )

        assert cvat_cli_config.create_base_command_string() == expected_base_command

    def test_cvat_annotation_handler_correct_base_command_from_password_file(self):
        cvat_cli_config = CVATCLIConfig(
            cli_path="CVAT_DIR/utils/cli/cli.py",
            server_host="http://localhost",
            server_port=8080,
            disable_ssl_verify=True,
            auth="USERNAME",
            password_path=os.path.join(
                self.project_root,
                "test_data/test_cvat_annotation_handler/cvat_password.txt",
            ),
        )

        expected_base_command = (
            f"{sys.executable} "
            "CVAT_DIR/utils/cli/cli.py "
            "--server-host http://localhost "
            "--server-port 8080 "
            "--auth USERNAME:PASSWORD"
        )

        assert cvat_cli_config.create_base_command_string() == expected_base_command

    def test_cvat_annotation_handler_correct_base_command_password_file_not_exist(self):
        cvat_cli_config = CVATCLIConfig(
            cli_path="CVAT_DIR/utils/cli/cli.py",
            server_host="http://localhost",
            server_port=8080,
            disable_ssl_verify=True,
            auth="USERNAME",
            password_path="NO_PATH",
        )

        expected_base_command = (
            f"{sys.executable} "
            "CVAT_DIR/utils/cli/cli.py "
            "--server-host http://localhost "
            "--server-port 8080 "
            "--auth USERNAME"
        )

        assert cvat_cli_config.create_base_command_string() == expected_base_command

    @pytest.mark.usefixtures("getpass_mock")
    def test_cvat_annotation_handler_correct_base_command_password_input(self):
        cvat_cli_config = CVATCLIConfig(
            cli_path="CVAT_DIR/utils/cli/cli.py",
            server_host="http://localhost",
            server_port=8080,
            disable_ssl_verify=True,
            auth="USERNAME",
        )

        expected_base_command = (
            f"{sys.executable} "
            "CVAT_DIR/utils/cli/cli.py "
            "--server-host http://localhost "
            "--server-port 8080 "
            "--auth USERNAME:PASSWORD"
        )

        assert cvat_cli_config.create_base_command_string() == expected_base_command

    def test_cvat_annotation_handler_correct_base_command_no_port(self):
        cvat_cli_config = CVATCLIConfig(
            cli_path="CVAT_DIR/utils/cli/cli.py",
            server_host="http://localhost",
            disable_ssl_verify=True,
            auth="USERNAME:PASSWORD",
        )

        expected_base_command = (
            f"{sys.executable} "
            "CVAT_DIR/utils/cli/cli.py "
            "--server-host http://localhost "
            "--auth USERNAME:PASSWORD"
        )

        assert cvat_cli_config.create_base_command_string() == expected_base_command

    def test_cvat_dump_cli_command(self):
        cvat_cli_config = CVATCLIConfig(
            cli_path="CVAT_DIR/utils/cli/cli.py",
            server_host="http://localhost",
            disable_ssl_verify=True,
            auth="USERNAME:PASSWORD",
        )

        assert CVATDumper._CVATDumper__create_cvat_dump_cli_command(
            base_command=cvat_cli_config.create_base_command_string(),
            target_zip_path="PATH_TO_ZIP/target.zip",
            task_info=CVATTaskInfoConfig(task_ID=0, annotation_format="PASCAL VOC 1.1"),
        ) == (
            f"{sys.executable} "
            "CVAT_DIR/utils/cli/cli.py "
            "--server-host http://localhost "
            "--auth USERNAME:PASSWORD "
            "dump 0 PATH_TO_ZIP/target.zip --format 'PASCAL VOC 1.1'"
        )

    def test_cvat_upload_cli_command(self):
        cvat_cli_config = CVATCLIConfig(
            cli_path="CVAT_DIR/utils/cli/cli.py",
            server_host="http://localhost",
            disable_ssl_verify=True,
            auth="USERNAME:PASSWORD",
        )

        assert PascalVOCUploader._PascalVOCUploader__create_cvat_upload_cli_command(
            base_command=cvat_cli_config.create_base_command_string(),
            target_zip_path="PATH_TO_ZIP/target.zip",
            task_info=CVATTaskInfoConfig(task_ID=0, annotation_format="PASCAL VOC 1.1"),
        ) == (
            f"{sys.executable} "
            "CVAT_DIR/utils/cli/cli.py "
            "--server-host http://localhost "
            "--auth USERNAME:PASSWORD "
            "upload 0 PATH_TO_ZIP/target.zip --format 'PASCAL VOC 1.1'"
        )

    def test_cvat_dumper_create_target_zip_path(self):
        assert CVATDumper._CVATDumper__create_target_zip_path(
            dump_task_config=CVATTaskDumpConfig(task_info=CVATTaskInfoConfig(task_ID=1))
        ) == os.path.join("tmp", "tmp-task.zip")

    @pytest.mark.usefixtures("subprocess_run_mock")
    def test_cvat_annotation_handler_download(self):
        cvat_annotation_handler = CVATAnnotationHandler(
            configuration=CVATAnnotationHandler.create_configuration(
                yaml_config_path=os.path.join(
                    self.project_root,
                    "test_data/test_cvat_annotation_handler/test_download/"
                    "test-download-task.yaml",
                ),
                string_replacement_map=self.string_replacement_map,
            )
        )

        if not os.path.isfile(
            cvat_annotation_handler.configuration.cvat_cli_config.password_path
        ):
            raise ValueError(
                "Please create the password file for your CVAT instance: %s"
                % cvat_annotation_handler.configuration.cvat_cli_config.password_path
            )

        cvat_annotation_handler.run()

    def test_cvat_annotation_handler_download_move_data(self):
        dataset_dir = os.path.join(self.project_root, "test_output", "dataset_dir")
        os.makedirs(dataset_dir, exist_ok=True)

        CVATDumper._CVATDumper__extract_and_copy_annotations_to_data_dir(
            target_zip_path=os.path.join(
                self.project_root,
                "test_data/test_cvat_annotation_handler/test_upload/mlcvzoo-test.zip",
            ),
            zip_extract_dir=os.path.join(self.project_root, "test_output"),
            dataset_dir=dataset_dir,
        )

    @pytest.mark.usefixtures("subprocess_run_failed_1_mock")
    def test_cvat_annotation_handler_download_failed(self):
        cvat_annotation_handler = CVATAnnotationHandler(
            configuration=CVATAnnotationHandler.create_configuration(
                yaml_config_path=os.path.join(
                    self.project_root,
                    "test_data/test_cvat_annotation_handler/test_download/"
                    "test-download-task.yaml",
                ),
                string_replacement_map=self.string_replacement_map,
            )
        )

        if not os.path.isfile(
            cvat_annotation_handler.configuration.cvat_cli_config.password_path
        ):
            raise ValueError(
                "Please create the password file for your CVAT instance: %s"
                % cvat_annotation_handler.configuration.cvat_cli_config.password_path
            )

        with self.assertRaises(OSError):
            cvat_annotation_handler.run()

    @pytest.mark.usefixtures("subprocess_run_mock")
    def test_cvat_annotation_handler_upload(self):
        cvat_annotation_handler = CVATAnnotationHandler(
            configuration=CVATAnnotationHandler.create_configuration(
                yaml_config_path=os.path.join(
                    self.project_root,
                    "test_data/test_cvat_annotation_handler/test_upload/"
                    "test-upload-task.yaml",
                ),
                string_replacement_map=self.string_replacement_map,
            )
        )

        if not os.path.isfile(
            cvat_annotation_handler.configuration.cvat_cli_config.password_path
        ):
            raise ValueError(
                "Please create the password file for your CVAT instance: %s"
                % cvat_annotation_handler.configuration.cvat_cli_config.password_path
            )

        cvat_annotation_handler.run()

        assert os.path.isfile(
            cvat_annotation_handler.configuration.upload_task_configs[0].target_zip_path
        )

        expected_extract_zip_dir = os.path.join(
            self.project_root,
            "test_output/test_cvat_annotation_handler/expected_extract_zip/",
        )
        test_extract_dir = os.path.join(
            self.project_root,
            "test_output/test_cvat_annotation_handler/test_extract_upload/",
        )

        extract_zip_data(
            zip_path=os.path.join(
                self.project_root,
                "test_data/test_cvat_annotation_handler/test_upload/mlcvzoo-test.zip",
            ),
            zip_extract_dir=expected_extract_zip_dir,
        )
        extract_zip_data(
            zip_path=cvat_annotation_handler.configuration.upload_task_configs[
                0
            ].target_zip_path,
            zip_extract_dir=test_extract_dir,
        )

        relative_test_path_1 = "Annotations/cars.xml"
        relative_test_path_2 = "Annotations/person.xml"
        relative_test_path_3 = "Annotations/truck.xml"

        assert _xml_equal(
            os.path.join(expected_extract_zip_dir, relative_test_path_1),
            os.path.join(test_extract_dir, relative_test_path_1),
        )

        assert _xml_equal(
            os.path.join(expected_extract_zip_dir, relative_test_path_2),
            os.path.join(test_extract_dir, relative_test_path_2),
        )

        assert _xml_equal(
            os.path.join(expected_extract_zip_dir, relative_test_path_3),
            os.path.join(test_extract_dir, relative_test_path_3),
        )

    @pytest.mark.usefixtures("subprocess_run_mock")
    def test_cvat_annotation_handler_upload_predictions(self):
        cvat_annotation_handler = CVATAnnotationHandler(
            configuration=CVATAnnotationHandler.create_configuration(
                yaml_config_path=os.path.join(
                    self.project_root,
                    "test_data/test_cvat_annotation_handler/test_upload/"
                    "test-upload-task-predictions.yaml",
                ),
                string_replacement_map=self.string_replacement_map,
            )
        )

        if not os.path.isfile(
            cvat_annotation_handler.configuration.cvat_cli_config.password_path
        ):
            raise ValueError(
                "Please create the password file for your CVAT instance: %s"
                % cvat_annotation_handler.configuration.cvat_cli_config.password_path
            )

        cvat_annotation_handler.run()

        assert os.path.isfile(
            cvat_annotation_handler.configuration.upload_task_configs[0].target_zip_path
        )

        test_extract_dir = os.path.join(
            self.project_root,
            "test_output/test_cvat_annotation_handler/test_extract_upload_predictions/",
        )

        extract_zip_data(
            zip_path=cvat_annotation_handler.configuration.upload_task_configs[
                0
            ].target_zip_path,
            zip_extract_dir=test_extract_dir,
        )

        assert _xml_equal(
            os.path.join(
                self.project_root,
                "test_data/annotations/pascal_voc/dummy_task/cars.xml",
            ),
            os.path.join(test_extract_dir, "Annotations/cars.xml"),
        )

        assert _xml_equal(
            os.path.join(
                self.project_root,
                "test_data/annotations/pascal_voc/dummy_task/person.xml",
            ),
            os.path.join(test_extract_dir, "Annotations/person.xml"),
        )

        assert _xml_equal(
            os.path.join(
                self.project_root,
                "test_data/annotations/pascal_voc/dummy_task/truck.xml",
            ),
            os.path.join(test_extract_dir, "Annotations/truck.xml"),
        )

    @pytest.mark.usefixtures("subprocess_run_mock")
    def test_cvat_annotation_handler_upload_predictions_only_create(self):
        cvat_annotation_handler = CVATAnnotationHandler(
            configuration=CVATAnnotationHandler.create_configuration(
                yaml_config_path=os.path.join(
                    self.project_root,
                    "test_data/test_cvat_annotation_handler/test_upload/"
                    "test-upload-task-predictions_only-create.yaml",
                ),
                string_replacement_map=self.string_replacement_map,
            )
        )

        if not os.path.isfile(
            cvat_annotation_handler.configuration.cvat_cli_config.password_path
        ):
            raise ValueError(
                "Please create the password file for your CVAT instance: %s"
                % cvat_annotation_handler.configuration.cvat_cli_config.password_path
            )

        cvat_annotation_handler.run()

        assert os.path.isfile(
            cvat_annotation_handler.configuration.upload_task_configs[0].target_zip_path
        )

        test_extract_dir = os.path.join(
            self.project_root,
            "test_output/test_cvat_annotation_handler/"
            "test_extract_upload_predictions_only_create/",
        )

        extract_zip_data(
            zip_path=cvat_annotation_handler.configuration.upload_task_configs[
                0
            ].target_zip_path,
            zip_extract_dir=test_extract_dir,
        )

        assert _xml_equal(
            os.path.join(
                self.project_root,
                "test_data/annotations/pascal_voc/dummy_task/cars.xml",
            ),
            os.path.join(test_extract_dir, "Annotations/cars.xml"),
        )

        assert _xml_equal(
            os.path.join(
                self.project_root,
                "test_data/annotations/pascal_voc/dummy_task/person.xml",
            ),
            os.path.join(test_extract_dir, "Annotations/person.xml"),
        )

        assert _xml_equal(
            os.path.join(
                self.project_root,
                "test_data/annotations/pascal_voc/dummy_task/truck.xml",
            ),
            os.path.join(test_extract_dir, "Annotations/truck.xml"),
        )

    @pytest.mark.usefixtures("subprocess_run_mock")
    def test_cvat_annotation_handler_download_and_upload(self):
        cvat_annotation_handler = CVATAnnotationHandler(
            configuration=CVATAnnotationHandler.create_configuration(
                yaml_config_path=os.path.join(
                    self.project_root,
                    "test_data/test_cvat_annotation_handler/"
                    "test-download-and-upload-task.yaml",
                ),
                string_replacement_map=self.string_replacement_map,
            )
        )

        if not os.path.isfile(
            cvat_annotation_handler.configuration.cvat_cli_config.password_path
        ):
            raise ValueError(
                "Please create the password file for your CVAT instance: %s"
                % cvat_annotation_handler.configuration.cvat_cli_config.password_path
            )

        cvat_annotation_handler.run()

    @pytest.mark.usefixtures("shutil_rmtree_mock", "os_path_isdir_mock")
    def test_extract_zip_data_remove_existing(self):
        extract_zip_data(
            zip_path=os.path.join(
                self.project_root,
                "test_data/test_cvat_annotation_handler/test_upload/mlcvzoo-test.zip",
            ),
            zip_extract_dir=os.path.join(
                self.project_root, "test_output", "zip_extract_dir"
            ),
        )

    @pytest.mark.usefixtures("os_path_isdir_mock")
    def test_extract_zip_data_skip(self):
        extract_zip_data(
            zip_path=os.path.join(
                self.project_root,
                "test_data/test_cvat_annotation_handler/test_upload/mlcvzoo-test.zip",
            ),
            zip_extract_dir=os.path.join(
                self.project_root, "test_output", "zip_extract_dir"
            ),
            remove_existing=False,
        )

    def test_cvat_annotation_handler_tool(self):
        command = "%s %s --replacement-config-path %s --log-level 'DEBUG'" % (
            os.path.join(
                self.project_root,
                "mlcvzoo_util/cvat_annotation_handler/cvat_annotation_handler.py",
            ),
            os.path.join(
                self.project_root,
                "test_data/test_cvat_annotation_handler/test_upload/"
                "test-upload-task-predictions_only-create.yaml",
            ),
            self._gen_replacement_config(),
        )

        command_split = [sys.executable]
        command_split.extend(shlex.split(command))

        result = subprocess.run(args=command_split, env=os.environ.copy(), check=False)

        assert result.returncode == 0

    def test_cvat_annotation_handler_too_wrong_parameter(self):
        command = "%s --wrong-parameter test" % (
            os.path.join(
                self.project_root,
                "mlcvzoo_util/mlcvzoo_util//cvat_annotation_handler.py",
            ),
        )

        command_split = [sys.executable]
        command_split.extend(shlex.split(command))

        result = subprocess.run(args=command_split, env=os.environ.copy(), check=False)

        assert result.returncode == 2

    def test_cvat_dumper_pre_cleanup(self):
        assert not CVATDumper._CVATDumper__pre_clean_up_and_determine_skip(
            overwrite_existing_zip=False,
            target_zip_path=os.path.join(
                self.project_root,
                "test_data/test_cvat_annotation_handler/test_download/NO_FILE.zip",
            ),
        )

    def test_cvat_dumper_pre_cleanup_not_overwrite_and_skip(self):
        assert CVATDumper._CVATDumper__pre_clean_up_and_determine_skip(
            overwrite_existing_zip=False,
            target_zip_path=os.path.join(
                self.project_root,
                "test_data/test_cvat_annotation_handler/test_download/mlcvzoo-test.zip",
            ),
        )

    def test_cvat_dumper_pre_cleanup_overwrite(self):
        target_zip_path = os.path.join(
            self.project_root,
            "test_data/test_cvat_annotation_handler/test_download/mlcvzoo-test.zip",
        )

        copy_target_zip_path = os.path.join(
            self.project_root,
            "test_data/test_cvat_annotation_handler/test_download/mlcvzoo-test_copy.zip",
        )

        shutil.copy(target_zip_path, copy_target_zip_path)

        assert not CVATDumper._CVATDumper__pre_clean_up_and_determine_skip(
            overwrite_existing_zip=True, target_zip_path=copy_target_zip_path
        )

        assert not os.path.isfile(copy_target_zip_path)

    def test_cvat_dumper_post_cleanup(self):
        target_zip_path = os.path.join(
            self.project_root,
            "test_data/test_cvat_annotation_handler/test_download/mlcvzoo-test.zip",
        )

        copy_target_zip_path = os.path.join(
            self.project_root,
            "test_data/test_cvat_annotation_handler/test_download/mlcvzoo-test_copy.zip",
        )

        zip_extract_dir = os.path.join(
            self.project_root,
            "test_data/test_cvat_annotation_handler/test_download/mlcvzoo-test/",
        )

        os.makedirs(zip_extract_dir, exist_ok=True)

        shutil.copy(target_zip_path, copy_target_zip_path)

        CVATDumper._CVATDumper__post_clean_up(
            target_zip_path=copy_target_zip_path, zip_extract_dir=zip_extract_dir
        )

        assert not os.path.isfile(copy_target_zip_path)
        assert not os.path.isdir(zip_extract_dir)

    def test_cvat_uploader_write_upload_zip_file_empty_target_zip_path(self):
        zip_path = os.path.join(
            self.project_root,
            "test_data/test_cvat_annotation_handler/test_download/mlcvzoo-test.zip",
        )

        target_zip_path = os.path.join(
            self.project_root,
            "test_data/test_cvat_annotation_handler/test_download/"
            "mlcvzoo-test_upload.zip",
        )

        shutil.copy(zip_path, target_zip_path)

        PascalVOCUploader.write_upload_zip_file(
            upload_task_config=CVATTaskUploadConfig(
                task_info=CVATTaskInfoConfig(task_ID=1),
                execute_upload=True,
                source_zip_path=zip_path,
                use_prediction_data=False,
                target_zip_path="",
                prediction_data_dir=os.path.join(
                    self.project_root, "test_data/annotations/pascal_voc/dummy_task"
                ),
            )
        )

        assert os.path.isfile(target_zip_path)

    def test_cvat_uploader_write_upload_zip_file_no_prediction_data_dir(self):
        zip_path = os.path.join(
            self.project_root,
            "test_data/test_cvat_annotation_handler/test_download/mlcvzoo-test.zip",
        )

        target_zip_path = os.path.join(
            self.project_root,
            "test_data/test_cvat_annotation_handler/test_download/"
            "mlcvzoo-test_upload.zip",
        )

        shutil.copy(zip_path, target_zip_path)

        with self.assertRaises(ValueError) as value_error:
            PascalVOCUploader.write_upload_zip_file(
                upload_task_config=CVATTaskUploadConfig(
                    task_info=CVATTaskInfoConfig(task_ID=1),
                    execute_upload=True,
                    source_zip_path=zip_path,
                    use_prediction_data=False,
                    target_zip_path=target_zip_path,
                    prediction_data_dir="",
                )
            )

            assert (
                str(value_error)
                == "prediction_data_dir='' does not exist! Please specify a "
                "directory in order to generatea prediction zip file for the upload to CVAT!"
            )

    def test_cvat_uploader_post_clean_up(self):
        zip_path = os.path.join(
            self.project_root,
            "test_data/test_cvat_annotation_handler/test_download/mlcvzoo-test.zip",
        )

        copy_source_zip_path = os.path.join(
            self.project_root,
            "test_data/test_cvat_annotation_handler/test_download/"
            "mlcvzoo-test_source-copy.zip",
        )

        copy_target_zip_path = os.path.join(
            self.project_root,
            "test_data/test_cvat_annotation_handler/test_download/"
            "mlcvzoo-test_target-copy.zip",
        )

        shutil.copy(zip_path, copy_source_zip_path)
        shutil.copy(zip_path, copy_target_zip_path)

        PascalVOCUploader._PascalVOCUploader__post_clean_up(
            source_zip_path=copy_source_zip_path, target_zip_path=copy_target_zip_path
        )

        assert not os.path.isfile(copy_source_zip_path)
        assert not os.path.isfile(copy_target_zip_path)

    def test_cvat_uploader_post_clean_up_no_files(self):
        PascalVOCUploader._PascalVOCUploader__post_clean_up(
            source_zip_path="source_zip_path", target_zip_path="target_zip_path"
        )

    def test_cvat_annotation_handler_failed_constructor(self):
        argv_copy = copy.deepcopy(sys.argv)

        sys.argv = [sys.argv[0]]

        with pytest.raises(SystemExit) as system_exit:
            cvat_annotation_handler_main()

        assert system_exit.value.code == 2

        sys.argv = argv_copy


if __name__ == "__main__":
    main()



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Module for the definition of a super class for any mlcvzoo module that is
providing a commandline interface.
"""

import argparse
import logging
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Generic, Optional, Type, TypeVar, cast

from config_builder import BaseConfigClass, ConfigBuilder

ConfigurationType = TypeVar("ConfigurationType", bound=BaseConfigClass)


class DictAction(argparse.Action):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super(DictAction, self).__init__(*args, **kwargs)
        self.nargs = "+"

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Any,
        option_string: Optional[Any] = None,
    ) -> None:
        # Get or create initial parameter value list
        current_values = getattr(namespace, self.dest, []) or []
        setattr(namespace, self.dest, current_values)

        # Get and store new parameter values
        parameter_values = getattr(namespace, self.dest)
        parameter_values.append(dict(v.split(":") for v in values))


class MLCVZooCLITool(ABC, Generic[ConfigurationType]):
    @staticmethod
    @abstractmethod
    def create_cli_configuration(
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> ConfigurationType:
        """
        Create a configuration of type ConfigurationType

        Expected implementation:

        return cast(
            YourConfigurationClass,
            YourCLITool._create_cli_configuration(
                configuration_class=YourConfigurationClass,
                string_replacement_map=string_replacement_map,
                configure_argparse=YourCLITool.configure_argparse,
            )
        )

        Args:
            string_replacement_map: A dictionary that defines placeholders which can be used
                                    while parsing the file. They can be understood as variables
                                    that can be used to define configs that are valid across
                                    multiple devices.
        Returns:
            The created configuration
        """
        raise NotImplementedError(
            "Must be implemented by sub-class: create_cli_configuration(...)."
        )

    @staticmethod
    def _create_cli_configuration(
        configuration_class: Type[ConfigurationType],
        string_replacement_map: Optional[Dict[str, str]] = None,
        configure_argparse: Optional[Callable[[argparse.ArgumentParser], None]] = None,
    ) -> ConfigurationType:
        return cast(
            ConfigurationType,
            ConfigBuilder(
                class_type=configuration_class,
                string_replacement_map=string_replacement_map,
                configure_argparse=configure_argparse,
                use_argparse=True,
            ).configuration,
        )

    @staticmethod
    def configure_argparse(
        parser: argparse.ArgumentParser,
    ) -> None:
        log_dir_option = "log-dir"
        log_level_option = "log-level"

        parser.add_argument(
            f"--{log_dir_option}",
            help="Directory where log-files should be written to",
            type=str,
        )

        parser.add_argument(
            f"--{log_level_option}",
            help="Define the logging level",
            type=str,
            default=logging.INFO,
            choices=logging._nameToLevel.keys(),
        )


def configure_model_argparse(
    parser: argparse.ArgumentParser,
) -> None:
    parser.add_argument(
        "--class-type",
        help="Class type of the model, respectively "
        "the string identifier of a model that is "
        "registered in the ModelRegistry of the mlcvzoo",
        type=str,
    )
    parser.add_argument(
        "--constructor-parameters",
        help="A list of key-value pairs "
        "in the format KEY:VALUE that should overwrite parameter of the model constructor",
        action=DictAction,
    )



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

import copy
import logging
import os
import sys
from typing import Dict, List, Optional, Tuple
from unittest import main
from unittest.mock import MagicMock

import mlflow
import related
from attr import define
from mlcvzoo_base.api.configuration import ModelConfiguration
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.interfaces import Classifiable, NetBased
from mlcvzoo_base.api.model import ConfigurationType, DataType, Model, PredictionType
from mlcvzoo_base.configuration.class_mapping_config import ClassMappingConfig
from mlcvzoo_base.evaluation.object_detection.data_classes import (
    ODMetrics,
    ODModelEvaluationMetrics,
)
from pytest import fixture, mark
from pytest_mock import MockerFixture
from test_template import TestTemplate

from mlcvzoo_util.model_evaluator.configuration import CheckpointConfig
from mlcvzoo_util.model_evaluator.metric_factory import MetricFactory
from mlcvzoo_util.model_evaluator.model_evaluator import ModelEvaluator
from mlcvzoo_util.model_evaluator.model_evaluator import (
    __get_model_checkpoint as get_model_checkpoint,
)
from mlcvzoo_util.model_evaluator.model_evaluator import main as model_evaluator_main
from mlcvzoo_util.model_evaluator.structs import CheckpointInfo

logger = logging.getLogger(__name__)

# fmt: off
test_evaluated_checkpoint_metrics = {
    'test_model_checkpoint': ODModelEvaluationMetrics(
        model_specifier='test_model',
        metrics_dict={
            0.5: {
                'ALL': {
                    '0_person': ODMetrics.from_dict({"TP": 1, "FP": 0, "FN": 0, "PR": 1.0, "RC": 1.0, "F1": 1.0, "AP": 1.0, "COUNT": 1}),
                    '1_truck': ODMetrics.from_dict({"TP": 1, "FP": 0, "FN": 0, "PR": 1.0, "RC": 1.0, "F1": 1.0, "AP": 1.0, "COUNT": 1}),
                    '2_car': ODMetrics.from_dict({"TP": 1, "FP": 0, "FN": 0, "PR": 1.0, "RC": 1.0, "F1": 1.0, "AP": 1.0, "COUNT": 1}),
                    '3_lp': ODMetrics.from_dict({"TP": 0, "FP": 0, "FN": 0, "PR": 0.0, "RC": 0.0, "F1": 0.0, "AP": 0.0, "COUNT": 0}),
                },
                'S': {
                    '0_person': ODMetrics.from_dict({"TP": 0, "FP": 0, "FN": 0, "PR": 0.0, "RC": 0.0, "F1": 0.0, "AP": 0.0, "COUNT": 0}),
                    '1_truck': ODMetrics.from_dict({"TP": 0, "FP": 0, "FN": 0, "PR": 0.0, "RC": 0.0, "F1": 0.0, "AP": 0.0, "COUNT": 0}),
                    '2_car': ODMetrics.from_dict({"TP": 0, "FP": 0, "FN": 0, "PR": 0.0, "RC": 0.0, "F1": 0.0, "AP": 0.0, "COUNT": 0}),
                    '3_lp': ODMetrics.from_dict({"TP": 0, "FP": 0, "FN": 0, "PR": 0.0, "RC": 0.0, "F1": 0.0, "AP": 0.0, "COUNT": 0}),
                },
                'M': {
                    '0_person': ODMetrics.from_dict({"TP": 0, "FP": 0, "FN": 0, "PR": 0.0, "RC": 0.0, "F1": 0.0, "AP": 0.0, "COUNT": 0}),
                    '1_truck': ODMetrics.from_dict({"TP": 0, "FP": 0, "FN": 0, "PR": 0.0, "RC": 0.0, "F1": 0.0, "AP": 0.0, "COUNT": 0}),
                    '2_car': ODMetrics.from_dict({"TP": 0, "FP": 0, "FN": 0, "PR": 0.0, "RC": 0.0, "F1": 0.0, "AP": 0.0, "COUNT": 0}),
                    '3_lp': ODMetrics.from_dict({"TP": 0, "FP": 0, "FN": 0, "PR": 0.0, "RC": 0.0, "F1": 0.0, "AP": 0.0, "COUNT": 0}),
                },
                'L': {
                    '0_person': ODMetrics.from_dict({"TP": 1, "FP": 0, "FN": 0, "PR": 1.0, "RC": 1.0, "F1": 1.0, "AP": 1.0, "COUNT": 1}),
                    '1_truck': ODMetrics.from_dict({"TP": 1, "FP": 0, "FN": 0, "PR": 1.0, "RC": 1.0, "F1": 1.0, "AP": 1.0, "COUNT": 1}),
                    '2_car': ODMetrics.from_dict({"TP": 1, "FP": 0, "FN": 0, "PR": 1.0, "RC": 1.0, "F1": 1.0, "AP": 1.0, "COUNT": 1}),
                    '3_lp': ODMetrics.from_dict({"TP": 0, "FP": 0, "FN": 0, "PR": 0.0, "RC": 0.0, "F1": 0.0, "AP": 0.0, "COUNT": 0})
                }
            }
        },
        metrics_image_info_dict={'0_person': {}, '1_truck': {}, '2_car': {}, '3_lp': {}}
    )
}
# fmt: on


@fixture(scope="function")
def select_evaluation_checkpoints_mock(mocker: MockerFixture) -> MagicMock:
    return mocker.patch(
        "mlcvzoo_util.model_evaluator.model_evaluator."
        "ModelEvaluator.select_evaluation_checkpoints",
        return_value=["TEST_CHECKPOINT", "TEST_CHECKPOINT_2"],
    )


@fixture(scope="function")
def restore_checkpoint_mock(mocker: MockerFixture) -> MagicMock:
    return mocker.patch(
        "mlcvzoo_util.model_evaluator.model_evaluator.ModelEvaluator._restore_checkpoint",
        return_value=None,
    )


@fixture(scope="function")
def set_inference_mock(mocker: MockerFixture) -> MagicMock:
    return mocker.patch(
        "mlcvzoo_base.configuration.model_config.ModelConfig.set_inference",
        return_value=None,
    )


@fixture(scope="function")
def determine_best_checkpoint_mock(mocker: MockerFixture) -> MagicMock:
    return mocker.patch(
        "mlcvzoo_util.model_evaluator.metric_factory.ODMetricFactory.determine_best_checkpoint",
        return_value=CheckpointInfo(
            path=MetricFactory.MODEL_STATE_INDICATOR,
            score=1.0,
        ),
    )


@fixture(scope="function")
def get_model_checkpoint_mock(mocker: MockerFixture) -> MagicMock:
    return mocker.patch(
        "mlcvzoo_util.model_evaluator.model_evaluator.__get_model_checkpoint",
        return_value=[
            os.path.join(
                "test_data/test_model_evaluator/test-checkpoint.pth",
            )
        ],
    )


@define
class TestConfig(ModelConfiguration):
    test_config: str = related.StringField(required=False, default="Hello World")


@fixture(scope="function")
def create_model_mock(mocker: MockerFixture) -> MagicMock:
    class TestModel(Model, Classifiable):
        def __init__(self):
            Model.__init__(
                self,
                configuration=TestConfig(unique_name="test-model"),
                init_for_inference=True,
            )
            Classifiable.__init__(
                self,
                mapper=AnnotationClassMapper(
                    class_mapping=ClassMappingConfig(
                        model_classes=[],
                        number_model_classes=0,
                    ),
                ),
            )

        @property
        def num_classes(self) -> int:
            return 0

        def get_classes_id_dict(self) -> Dict[int, str]:
            pass

        def predict(self, data_item: DataType) -> Tuple[DataType, List[PredictionType]]:
            pass

        @staticmethod
        def create_configuration(
            from_yaml: Optional[str] = None,
            configuration: Optional[ConfigurationType] = None,
            string_replacement_map: Optional[Dict[str, str]] = None,
        ) -> ConfigurationType:
            pass

    return mocker.patch(
        "mlcvzoo_util.model_evaluator.model_evaluator.ModelEvaluator.create_model",
        return_value=TestModel(),
    )


class TestModelEvaluator(TestTemplate):
    def test_metric_factory_base(self) -> None:
        with self.assertRaises(NotImplementedError):
            MetricFactory.log_results(
                checkpoint_log_mode="all",
                evaluated_checkpoint_metrics=test_evaluated_checkpoint_metrics,
                best_checkpoint=CheckpointInfo(
                    path="test_model_checkpoint",
                    score=1.0,
                ),
                logging_configs=None,
            )
        with self.assertRaises(NotImplementedError):
            MetricFactory.log_results(
                checkpoint_log_mode="all",
                evaluated_checkpoint_metrics=test_evaluated_checkpoint_metrics,
                best_checkpoint=CheckpointInfo(
                    path="test_model_checkpoint",
                    score=1.0,
                ),
                logging_configs=None,
            )

    @mark.usefixtures(
        "select_evaluation_checkpoints_mock",
        "restore_checkpoint_mock",
        "set_inference_mock",
    )
    def test_model_evaluator_main(self):
        argv_copy = copy.deepcopy(sys.argv)

        sys.argv = [sys.argv[0]]

        sys.argv.extend(
            [
                os.path.join(
                    self.project_root,
                    "test_data/test_model_evaluator/"
                    "test_model_evaluator_without_mlflow.yaml",
                ),
                "--replacement-config-path",
                self._gen_replacement_config(),
                "--log-dir",
                os.path.join(self.project_root, "test_output", "logs"),
                "--log-level",
                "DEBUG",
            ]
        )

        model_evaluator_main()

        sys.argv = argv_copy

    @mark.usefixtures(
        "select_evaluation_checkpoints_mock",
        "restore_checkpoint_mock",
        "set_inference_mock",
        "create_model_mock",
    )
    def test_model_evaluator_wrong_model(self):
        argv_copy = copy.deepcopy(sys.argv)

        sys.argv = [sys.argv[0]]

        sys.argv.extend(
            [
                os.path.join(
                    self.project_root,
                    "test_data/test_model_evaluator/"
                    "test_model_evaluator_without_mlflow.yaml",
                ),
                "--replacement-config-path",
                self._gen_replacement_config(),
                "--log-dir",
                os.path.join(self.project_root, "test_output", "logs"),
                "--log-level",
                "DEBUG",
            ]
        )

        with self.assertRaises(ValueError):
            model_evaluator_main()

        sys.argv = argv_copy

    @mark.usefixtures(
        "select_evaluation_checkpoints_mock",
        "restore_checkpoint_mock",
        "set_inference_mock",
    )
    def test_model_evaluator_main_single_mode(self):
        argv_copy = copy.deepcopy(sys.argv)

        sys.argv = [sys.argv[0]]

        sys.argv.extend(
            [
                os.path.join(
                    self.project_root,
                    "test_data/test_model_evaluator/"
                    "test_model_evaluator_without_mlflow.yaml",
                ),
                "--replacement-config-path",
                self._gen_replacement_config(),
                "--log-dir",
                os.path.join(self.project_root, "test_output", "logs"),
                "--log-level",
                "DEBUG",
                "--single-mode",
            ]
        )

        model_evaluator_main()

        sys.argv = argv_copy

    @mark.usefixtures(
        "select_evaluation_checkpoints_mock",
        "restore_checkpoint_mock",
        "set_inference_mock",
    )
    def test_model_evaluator_main_single_mode_from_argparse(self):
        argv_copy = copy.deepcopy(sys.argv)

        sys.argv = [sys.argv[0]]

        from_yaml = os.path.join(
            self.project_root,
            "test_data/test_ReadFromFileObjectDetectionModel/"
            "read-from-file_pascal_voc_clean.yaml",
        )

        sys.argv.extend(
            [
                os.path.join(
                    self.project_root,
                    "test_data/test_model_evaluator/"
                    "test_model_evaluator_without_mlflow.yaml",
                ),
                "--constructor-parameters",
                f"from_yaml:{from_yaml}",
                "--replacement-config-path",
                self._gen_replacement_config(),
                "--log-dir",
                os.path.join(self.project_root, "test_output", "logs"),
                "--log-level",
                "DEBUG",
                "--single-mode",
            ]
        )

        model_evaluator_main()

        sys.argv = argv_copy

    @mark.usefixtures(
        "select_evaluation_checkpoints_mock",
        "restore_checkpoint_mock",
        "set_inference_mock",
    )
    def test_model_evaluator_main_single_mode_with_tb(self):
        mlflow.end_run()

        argv_copy = copy.deepcopy(sys.argv)

        sys.argv = [sys.argv[0]]

        sys.argv.extend(
            [
                os.path.join(
                    self.project_root,
                    "test_data/test_model_evaluator/"
                    "test_model_evaluator_tensorboard.yaml",
                ),
                "--replacement-config-path",
                self._gen_replacement_config(),
                "--log-dir",
                os.path.join(self.project_root, "test_output", "logs"),
                "--log-level",
                "DEBUG",
                "--single-mode",
            ]
        )

        model_evaluator_main()

        sys.argv = argv_copy

    @mark.usefixtures(
        "select_evaluation_checkpoints_mock",
        "restore_checkpoint_mock",
        "set_inference_mock",
    )
    def test_model_evaluator_main_single_mode_with_mlflow(self):
        argv_copy = copy.deepcopy(sys.argv)

        sys.argv = [sys.argv[0]]

        sys.argv.extend(
            [
                os.path.join(
                    self.project_root,
                    "test_data/test_model_evaluator/"
                    "test_model_evaluator_with_mlflow.yaml",
                ),
                "--replacement-config-path",
                self._gen_replacement_config(),
                "--log-dir",
                os.path.join(self.project_root, "test_output", "logs"),
                "--log-level",
                "DEBUG",
                "--single-mode",
            ]
        )

        model_evaluator_main()

        sys.argv = argv_copy

    @mark.usefixtures(
        "select_evaluation_checkpoints_mock",
        "restore_checkpoint_mock",
        "set_inference_mock",
    )
    def test_model_evaluator_main_single_mode_with_mlflow_all_checkpoints(self):
        argv_copy = copy.deepcopy(sys.argv)

        sys.argv = [sys.argv[0]]

        sys.argv.extend(
            [
                os.path.join(
                    self.project_root,
                    "test_data/test_model_evaluator/"
                    "test_model_evaluator_with_mlflow_all-checkpoint.yaml",
                ),
                "--replacement-config-path",
                self._gen_replacement_config(),
                "--log-dir",
                os.path.join(self.project_root, "test_output", "logs"),
                "--log-level",
                "DEBUG",
                "--single-mode",
            ]
        )

        model_evaluator_main()

        sys.argv = argv_copy

    @mark.usefixtures(
        "select_evaluation_checkpoints_mock",
        "restore_checkpoint_mock",
        "set_inference_mock",
        "determine_best_checkpoint_mock",
    )
    def test_model_evaluator_main_single_mode_with_mlflow_best_checkpoint(self):
        argv_copy = copy.deepcopy(sys.argv)

        sys.argv = [sys.argv[0]]

        sys.argv.extend(
            [
                os.path.join(
                    self.project_root,
                    "test_data/test_model_evaluator/"
                    "test_model_evaluator_with_mlflow_best-checkpoint.yaml",
                ),
                "--replacement-config-path",
                self._gen_replacement_config(),
                "--log-dir",
                os.path.join(self.project_root, "test_output", "logs"),
                "--log-level",
                "DEBUG",
                "--single-mode",
            ]
        )

        model_evaluator_main()

        sys.argv = argv_copy

    @mark.usefixtures(
        "select_evaluation_checkpoints_mock",
        "restore_checkpoint_mock",
        "set_inference_mock",
    )
    def test_model_evaluator_main_single_mode_with_mlflow_wrong_checkpoint(self):
        argv_copy = copy.deepcopy(sys.argv)

        sys.argv = [sys.argv[0]]

        sys.argv.extend(
            [
                os.path.join(
                    self.project_root,
                    "test_data/test_model_evaluator/"
                    "test_model_evaluator_with_mlflow_wrong-checkpoint.yaml",
                ),
                "--replacement-config-path",
                self._gen_replacement_config(),
                "--log-dir",
                os.path.join(self.project_root, "test_output", "logs"),
                "--log-level",
                "DEBUG",
                "--single-mode",
            ]
        )

        with self.assertRaises(ValueError):
            model_evaluator_main()

        sys.argv = argv_copy

    @mark.usefixtures(
        "select_evaluation_checkpoints_mock",
        "restore_checkpoint_mock",
        "set_inference_mock",
    )
    def test_model_evaluator_main_single_mode_with_mlflow_wrong_model(self):
        argv_copy = copy.deepcopy(sys.argv)

        sys.argv = [sys.argv[0]]

        sys.argv.extend(
            [
                os.path.join(
                    self.project_root,
                    "test_data/test_model_evaluator/"
                    "test_model_evaluator_with_mlflow_wrong-model.yaml",
                ),
                "--replacement-config-path",
                self._gen_replacement_config(),
                "--log-dir",
                os.path.join(self.project_root, "test_output", "logs"),
                "--log-level",
                "DEBUG",
                "--single-mode",
            ]
        )

        with self.assertRaises(ValueError):
            model_evaluator_main()

        sys.argv = argv_copy

    @mark.usefixtures(
        "select_evaluation_checkpoints_mock",
        "restore_checkpoint_mock",
        "set_inference_mock",
    )
    def test_model_evaluator_no_object_detection_model(self):
        argv_copy = copy.deepcopy(sys.argv)

        sys.argv = [sys.argv[0]]

        sys.argv.extend(
            [
                os.path.join(
                    self.project_root,
                    "test_data/test_model_evaluator/"
                    "test_model_evaluator_no_object_detection_model.yaml",
                ),
                "--replacement-config-path",
                self._gen_replacement_config(),
                "--log-dir",
                os.path.join(self.project_root, "test_output", "logs"),
                "--log-level",
                "DEBUG",
                "--single-mode",
            ]
        )

        with self.assertRaises(ValueError):
            model_evaluator_main()

        sys.argv = argv_copy

    @mark.usefixtures(
        "select_evaluation_checkpoints_mock",
        "restore_checkpoint_mock",
        "set_inference_mock",
        "get_model_checkpoint_mock",
    )
    def test_model_evaluator_get_model_checkpoint(self):
        argv_copy = copy.deepcopy(sys.argv)

        sys.argv = [sys.argv[0]]

        sys.argv.extend(
            [
                os.path.join(
                    self.project_root,
                    "test_data/test_model_evaluator/"
                    "test_model_evaluator_with_mlflow_best-checkpoint.yaml",
                ),
                "--replacement-config-path",
                self._gen_replacement_config(),
                "--log-dir",
                os.path.join(self.project_root, "test_output", "logs"),
                "--log-level",
                "DEBUG",
                "--single-mode",
            ]
        )

        model_evaluator_main()

        sys.argv = argv_copy

    @mark.usefixtures(
        "select_evaluation_checkpoints_mock",
        "restore_checkpoint_mock",
        "set_inference_mock",
    )
    def test_model_evaluator_no_factory(self):
        argv_copy = copy.deepcopy(sys.argv)

        sys.argv = [sys.argv[0]]

        sys.argv.extend(
            [
                os.path.join(
                    self.project_root,
                    "test_data/test_model_evaluator/"
                    "test_model_evaluator_no_factory.yaml",
                ),
                "--replacement-config-path",
                self._gen_replacement_config(),
                "--log-dir",
                os.path.join(self.project_root, "test_output", "logs"),
                "--log-level",
                "DEBUG",
            ]
        )

        with self.assertRaises(ValueError):
            model_evaluator_main()

        sys.argv = argv_copy

    def test_select_evaluation_checkpoints(self) -> None:
        assert ModelEvaluator.select_evaluation_checkpoints(
            checkpoint_config=CheckpointConfig(
                checkpoint_dir=os.path.join(self.project_root, "test_data/checkpoints"),
                checkpoint_filename_suffix=".pth",
            )
        ) == [
            os.path.join(self.project_root, "test_data/checkpoints/model.pth"),
            os.path.join(self.project_root, "test_data/checkpoints/model_2.pth"),
        ]

    def test_restore_checkpoint(self) -> None:
        class TestModel(Model, NetBased):
            def __init__(self):
                pass

            def predict(
                self, data_item: DataType
            ) -> Tuple[DataType, List[PredictionType]]:
                pass

            def store(self, checkpoint_path: str) -> None:
                pass

            def get_checkpoint_filename_suffix(self) -> str:
                return ""

            @staticmethod
            def create_configuration(
                from_yaml: Optional[str] = None,
                configuration: Optional[ConfigurationType] = None,
                string_replacement_map: Optional[Dict[str, str]] = None,
            ) -> ConfigurationType:
                pass

            def restore(self, checkpoint_path: str) -> None:
                pass

        ModelEvaluator._restore_checkpoint(
            checkpoint_path=os.path.join(
                self.project_root, "test_data/checkpoints/model.pth"
            ),
            model=TestModel(),
        )

    def test_restore_checkpoint_not_net_based(self) -> None:
        class TestModel(Model):
            def __init__(self):
                pass

            def predict(
                self, data_item: DataType
            ) -> Tuple[DataType, List[PredictionType]]:
                pass

            @staticmethod
            def create_configuration(
                from_yaml: Optional[str] = None,
                configuration: Optional[ConfigurationType] = None,
                string_replacement_map: Optional[Dict[str, str]] = None,
            ) -> ConfigurationType:
                pass

        with self.assertRaises(ValueError):
            ModelEvaluator._restore_checkpoint(
                checkpoint_path=os.path.join(
                    self.project_root, "test_data/checkpoints/model.pth"
                ),
                model=TestModel(),
            )

    def test_restore_checkpoint_wrong_path(self) -> None:
        class TestModel(Model):
            def __init__(self):
                pass

            def predict(
                self, data_item: DataType
            ) -> Tuple[DataType, List[PredictionType]]:
                pass

            @staticmethod
            def create_configuration(
                from_yaml: Optional[str] = None,
                configuration: Optional[ConfigurationType] = None,
                string_replacement_map: Optional[Dict[str, str]] = None,
            ) -> ConfigurationType:
                pass

        with self.assertRaises(ValueError):
            ModelEvaluator._restore_checkpoint(
                checkpoint_path=os.path.join(
                    self.project_root, "test_data/checkpoints/model.pt"
                ),
                model=TestModel(),
            )

    def test_model_evaluator_create_configuration(self) -> None:
        assert (
            ModelEvaluator.create_configuration(
                yaml_config_path=os.path.join(
                    self.project_root,
                    "test_data/test_model_evaluator/" "test_model_evaluator.yaml",
                ),
                string_replacement_map=self.string_replacement_map,
                no_checks=True,
            )
            is not None
        )

    def test_model_evaluator_get_model_checkpoint(self) -> None:
        class TestCheckpointConfig:
            def __init__(self) -> None:
                self.checkpoint_path: str = "test_checkpoint.pth"

        class TestInferenceConfig:
            def __init__(self) -> None:
                self.inference_config: TestCheckpointConfig = TestCheckpointConfig()

        class TestModel(Model, NetBased):
            def __init__(self) -> None:
                self.configuration = TestInferenceConfig()

            def predict(
                self, data_item: DataType
            ) -> Tuple[DataType, List[PredictionType]]:
                pass

            def store(self, checkpoint_path: str) -> None:
                pass

            def get_checkpoint_filename_suffix(self) -> str:
                return ""

            @staticmethod
            def create_configuration(
                from_yaml: Optional[str] = None,
                configuration: Optional[ConfigurationType] = None,
                string_replacement_map: Optional[Dict[str, str]] = None,
            ) -> ConfigurationType:
                pass

            def restore(self, checkpoint_path: str) -> None:
                pass

        assert get_model_checkpoint(model=TestModel()) == ["test_checkpoint.pth"]

    def test_model_evaluator_get_model_checkpoint_not_netbase(self) -> None:
        class TestModel(Model):
            def __init__(self) -> None:
                pass

            def predict(
                self, data_item: DataType
            ) -> Tuple[DataType, List[PredictionType]]:
                pass

            @staticmethod
            def create_configuration(
                from_yaml: Optional[str] = None,
                configuration: Optional[ConfigurationType] = None,
                string_replacement_map: Optional[Dict[str, str]] = None,
            ) -> ConfigurationType:
                pass

        assert get_model_checkpoint(model=TestModel()) is None

    def test_model_evaluator_get_model_checkpoint_accept_no_config(self) -> None:
        class TestModel(Model, NetBased):
            def __init__(self) -> None:
                pass

            def get_checkpoint_filename_suffix(self) -> str:
                return ""

            def predict(
                self, data_item: DataType
            ) -> Tuple[DataType, List[PredictionType]]:
                pass

            def store(self, checkpoint_path: str) -> None:
                pass

            @staticmethod
            def create_configuration(
                from_yaml: Optional[str] = None,
                configuration: Optional[ConfigurationType] = None,
                string_replacement_map: Optional[Dict[str, str]] = None,
            ) -> ConfigurationType:
                pass

            def restore(self, checkpoint_path: str) -> None:
                pass

        assert get_model_checkpoint(model=TestModel()) is None


if __name__ == "__main__":
    main()



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Module for the definition of a generic MetricFactory which provides
interface methods. These interfaces are consumed by the ModelEvaluator
in order to implement a generic metric computation functionally.
"""

import logging
from abc import ABC
from typing import Any, Dict, Generic, List, Optional, Type, TypeVar

import mlflow
from config_builder import BaseConfigClass
from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.model import Model, ObjectDetectionModel
from mlcvzoo_base.evaluation.object_detection.configuration import (
    TensorboardLoggingConfig,
)
from mlcvzoo_base.evaluation.object_detection.data_classes import (
    ODModelEvaluationMetrics,
)
from mlcvzoo_base.evaluation.object_detection.metrics_computation import (
    MetricsComputation,
)
from mlcvzoo_base.evaluation.object_detection.metrics_logging import (
    log_false_positive_info_to_tb,
    log_od_metrics_to_mlflow,
)
from mlcvzoo_base.evaluation.object_detection.model_evaluation import (
    evaluate_with_model,
)
from mlcvzoo_base.evaluation.object_detection.utils import generate_metric_table

from mlcvzoo_util.model_evaluator.configuration import ModelEvaluatorConfig
from mlcvzoo_util.model_evaluator.structs import CheckpointInfo, CheckpointLoggingModes

ModelType = TypeVar("ModelType", bound=Model[Any, Any, Any])
EvaluationMetricType = TypeVar("EvaluationMetricType")

logger = logging.getLogger(__name__)


class MetricFactory(ABC, Generic[ModelType, EvaluationMetricType]):
    """
    Super class for defining interfaces for metric computations.
    """

    # This constant is used to indicate that an entry in the evaluated_checkpoint_metrics
    # is not produced by loading a dedicated checkpoint of a model, but that the evaluation
    # have been executed by using the raw model object.
    MODEL_STATE_INDICATOR: str = "model_state"

    @staticmethod
    def compute_metrics(
        inference_model: ModelType,
        gt_annotations: List[BaseAnnotation],
        model_evaluator_config: ModelEvaluatorConfig,
    ) -> EvaluationMetricType:
        raise NotImplementedError(
            "Must be implemented by sub-class: compute_metrics(...)"
        )

    @staticmethod
    def determine_best_checkpoint(
        evaluated_checkpoint_metrics: Dict[str, EvaluationMetricType],
    ) -> CheckpointInfo:
        raise NotImplementedError(
            "Must be implemented by sub-class: determine_best_checkpoint(...)"
        )

    @staticmethod
    def log_results(
        checkpoint_log_mode: str,
        evaluated_checkpoint_metrics: Dict[str, EvaluationMetricType],
        best_checkpoint: CheckpointInfo,
        logging_configs: Optional[List[BaseConfigClass]] = None,
    ) -> None:
        raise NotImplementedError("Must be implemented by sub-class: log_results(...)")


class ODMetricFactory(
    MetricFactory[ObjectDetectionModel[Any, Any], ODModelEvaluationMetrics]
):
    """
    Implements the MetricFactory in order to provide an
    generic evaluation of ObjectDetectionModels.
    """

    @staticmethod
    def compute_metrics(
        inference_model: ObjectDetectionModel[Any, Any],
        gt_annotations: List[BaseAnnotation],
        model_evaluator_config: ModelEvaluatorConfig,
    ) -> ODModelEvaluationMetrics:
        return evaluate_with_model(
            model=inference_model,
            gt_annotations=gt_annotations,
            iou_thresholds=model_evaluator_config.iou_thresholds,
        )

    @staticmethod
    def determine_best_checkpoint(
        evaluated_checkpoint_metrics: Dict[str, ODModelEvaluationMetrics],
    ) -> CheckpointInfo:
        """
        Determine the best checkpoint based on the given overall AP metric per checkpoint.

        Returns:
            A CheckpointInfo object stating the best checkpoint
        """
        best_checkpoint = CheckpointInfo(path="", score=-1.0)

        for ckpt, model_metrics in evaluated_checkpoint_metrics.items():
            current_map = MetricsComputation.compute_average_ap(
                model_metrics=model_metrics
            )

            if current_map > best_checkpoint.score:
                best_checkpoint = CheckpointInfo(path=ckpt, score=current_map)

        return best_checkpoint

    @staticmethod
    def __log_checkpoint(
        checkpoint_log_mode: str,
        evaluated_checkpoint_metrics: Dict[str, ODModelEvaluationMetrics],
        best_checkpoint: CheckpointInfo,
    ) -> None:
        if not mlflow.active_run():
            logger.warning(
                "No mlflow run is active, logging of checkpoint(s) as artifacts "
                "will not take place"
            )
            return

        checkpoint_paths_to_log: List[str] = []

        if checkpoint_log_mode.lower() == CheckpointLoggingModes.ALL.value:
            checkpoint_paths_to_log.extend(evaluated_checkpoint_metrics.keys())

        elif checkpoint_log_mode.lower() == CheckpointLoggingModes.BEST.value:
            checkpoint_paths_to_log.append(best_checkpoint.path)

        elif checkpoint_log_mode.lower() == CheckpointLoggingModes.NONE.value:
            pass
        else:
            raise ValueError(
                f"The specified value for parameter: "
                f"'{checkpoint_log_mode}' "
                f"'checkpoint_log_mode' is invalid! Only 'all' and 'best' is allowed!"
            )

        for checkpoint_path in checkpoint_paths_to_log:
            if checkpoint_path != MetricFactory.MODEL_STATE_INDICATOR:
                mlflow.log_artifact(checkpoint_path)

    @staticmethod
    def log_results(
        checkpoint_log_mode: str,
        evaluated_checkpoint_metrics: Dict[str, ODModelEvaluationMetrics],
        best_checkpoint: CheckpointInfo,
        logging_configs: Optional[List[BaseConfigClass]] = None,
    ) -> None:
        """
        Logs evaluated metrics, checkpoints and parameters to mlflow as
        specified in configuration file.

        Returns:
            None
        """
        logger.info(f"Log results after evaluation")

        logger.info(
            generate_metric_table(
                metrics_dict=evaluated_checkpoint_metrics[
                    best_checkpoint.path
                ].metrics_dict,
                iou_threshold=list(
                    evaluated_checkpoint_metrics[
                        best_checkpoint.path
                    ].metrics_dict.keys()
                )[0],
            ).table
        )

        if logging_configs:
            for logging_config in logging_configs:
                if isinstance(logging_config, TensorboardLoggingConfig):
                    log_false_positive_info_to_tb(
                        model_name=evaluated_checkpoint_metrics[
                            best_checkpoint.path
                        ].model_specifier,
                        metric_image_info_dict=evaluated_checkpoint_metrics[
                            best_checkpoint.path
                        ].metrics_image_info_dict,
                        tb_logging_config=logging_config,
                    )

        # TODO: Add feature that determines the epoch from a given checkpoint path.
        #       Use it to fill the step parameter correctly.
        for step, (ckpt, metrics) in enumerate(evaluated_checkpoint_metrics.items()):
            for iou in metrics.metrics_dict:
                log_od_metrics_to_mlflow(
                    model_specifier=metrics.model_specifier,
                    metrics_dict=metrics.metrics_dict,
                    iou_threshold=float(iou),
                    step=step,
                )

        ODMetricFactory.__log_checkpoint(
            checkpoint_log_mode=checkpoint_log_mode,
            evaluated_checkpoint_metrics=evaluated_checkpoint_metrics,
            best_checkpoint=best_checkpoint,
        )

        logger.debug(
            f"Logged evaluated checkpoints with mode='%s'" % checkpoint_log_mode
        )


__metric_factory_dict: Dict[
    Type[Model[Any, Any, Any]], Type[MetricFactory[Any, Any]]
] = {
    ObjectDetectionModel: ODMetricFactory,  # type: ignore[type-abstract]
}


def get_factory(
    inference_model: Model[Any, Any, Any]
) -> Optional[Type[MetricFactory[Any, Any]]]:
    for key, value in __metric_factory_dict.items():
        if isinstance(inference_model, key):
            return value

    return None



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module for handling python logging"""

import logging
import os
import re
import sys
from datetime import datetime
from logging import StreamHandler
from logging.handlers import TimedRotatingFileHandler
from sys import stdout
from typing import List, Optional, Union

from mlcvzoo_base.utils import ensure_dir

logger = logging.getLogger(__name__)


class Logger:
    """Class for initializing python logging"""

    is_instantiated: bool = False

    log_handlers: List[Union[TimedRotatingFileHandler, StreamHandler]] = []  # type: ignore

    @staticmethod
    def init_logging_basic(
        log_dir: Optional[str] = None,
        log_file_postfix: Optional[str] = None,
        no_stdout: bool = False,
        root_log_level: Union[str, int] = logging.INFO,
    ) -> None:
        """
        Initializes the python logging. Per default only logging to stdout is used.
        When a log_dir is given, also an TimedRotatingFileHandler is added

        Args:
            log_dir: String, directory where the log-file is about to be stored
            log_file_postfix: String, postfix of log-file
            no_stdout: Bool, whether a standard output should be done
            root_log_level: String or int, defines root logging level

        Returns:
            None
        """

        root_log_level = Logger.get_log_level(log_level=root_log_level)

        if not Logger.is_instantiated:
            if log_dir is not None:
                if log_file_postfix is not None:
                    log_file_name = (
                        f"{log_file_postfix}_"
                        f"{datetime.now().strftime('%Y-%m-%dT_%H-%M')}.log"
                    )
                else:
                    log_file_name = (
                        f"log_{datetime.now().strftime('%Y-%m-%dT_%H-%M')}.log"
                    )

                log_file_path = os.path.join(log_dir, log_file_name)

                ensure_dir(log_file_path, verbose=True)

                time_rotating_file_handler = TimedRotatingFileHandler(
                    filename=log_file_path,
                    when="h",
                    interval=6,
                )
                time_rotating_file_handler.suffix = "%Y-%m-%d_%H.log"
                time_rotating_file_handler.extMatch = re.compile(
                    r"^\d{4}-\d{2}-\d{2}_\d{2}(\.\w+)?\.log$", re.ASCII
                )

                Logger.log_handlers.append(time_rotating_file_handler)

            if not no_stdout:
                stdout_handler = StreamHandler(stream=stdout)
                Logger.log_handlers.append(stdout_handler)

            logging.basicConfig(
                level=root_log_level,
                format="%(asctime)s %(levelname)-7.7s: "
                "[%(name)-30.30s]"
                "[%(threadName)-11.11s]"
                "[%(funcName)s():%(lineno)s] "
                "%(message)s",
                handlers=Logger.log_handlers,
            )

            Logger.is_instantiated = True
        else:
            logger.warning(
                "Logger has already been instantiated. "
                "Another call won't have any effect",
            )

    @staticmethod
    def get_log_level(log_level: Optional[Union[str, int]]) -> Union[str, int]:
        """
        Matches the correct python logging level for the
        given log_level.

        Args:
            log_level: The log level to match

        Returns:
            The matched logging level
        """
        if log_level is None:
            return logging.INFO

        _log_level: Union[str, int] = logging.getLevelName(log_level)

        if isinstance(_log_level, str) and "Level " in _log_level:
            return logging.INFO

        return _log_level



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Module for extracting images from a video and converting annotations from a Label Studio video
object detection/tracking task to annotations for a object detection task.
"""
from __future__ import annotations

import argparse
import json
import os
import sys
import zipfile
from collections import defaultdict
from dataclasses import dataclass
from tempfile import TemporaryDirectory
from typing import Any, Dict, List

import cv2


class LabelStudioTrackingTaskConverter:  # pylint: disable=R0903
    """Class to handle conversion of a Label Studio tracking task to a object detection task."""

    @dataclass
    class __Annotation:
        """
        Class for storing an annotation.
        """

        x_min: float
        y_min: float
        width: float
        height: float
        rotation: float
        labels: List[str]

    @dataclass
    class __VideoSpecs:
        """
        Class for storing video attributes.
        """

        width: int
        height: int
        frame_count: int
        frame_rate: float

    def __init__(
        self,
        video_path: str,
        tracking_annotations_path: str,
        image_archive_path: str,
        annotations_path: str,
    ) -> None:
        """
        Create a LabelStudioTrackingTaskConverter

        Args:
            video_path: Path to the video.
            tracking_annotations_path: Path to the Label Studios tracking annotations.
            image_archive_path: Path to write the archive with separate images to.
            annotations_path: Path to write the annotations for the detection task to.

        Returns:
            The created LabelStudioTrackingTaskConverter instance
        """

        if not os.path.exists(video_path):
            raise ValueError(f"Path to video '{video_path}' does not exist.")
        if not os.path.exists(tracking_annotations_path):
            raise ValueError(
                f"Path to tracking annotations '{tracking_annotations_path}' does not exist."
            )

        self.video_path: str = video_path
        self.tracking_annotations_path: str = tracking_annotations_path
        self.image_archive_path: str = image_archive_path
        self.annotations_path: str = annotations_path

        self.video_specs = self._parse_video_specs()
        self.annotations: Dict[
            int, List[LabelStudioTrackingTaskConverter.__Annotation]
        ] = defaultdict(list)

    def run(self) -> None:
        """
        Extract images from video and convert annotations.
        """

        self._parse_tracking_annotations_file()
        self._extract_images()
        self._save_annotations()

    def _parse_video_specs(self) -> LabelStudioTrackingTaskConverter.__VideoSpecs:
        """
        Parse attributes for a video file.

        Returns:
            The parsed video attributes.
        """

        video = cv2.VideoCapture(self.video_path)

        if video.isOpened():
            video_specs = LabelStudioTrackingTaskConverter.__VideoSpecs(
                width=int(video.get(cv2.CAP_PROP_FRAME_WIDTH)),
                height=int(video.get(cv2.CAP_PROP_FRAME_HEIGHT)),
                frame_count=int(video.get(cv2.CAP_PROP_FRAME_COUNT)),
                frame_rate=video.get(cv2.CAP_PROP_FPS),
            )
            video.release()

            return video_specs
        else:
            raise RuntimeError(f"Could not open video '{self.video_path}'")

    def _find_task(self) -> Dict[str, Any]:
        """
        Finds the task inside the tracking annotation file using the videos filename.
        """
        video_name = os.path.basename(self.video_path)

        with open(file=self.tracking_annotations_path, encoding="utf-8") as file:
            data = json.load(fp=file)

        if len(data) == 0:
            raise ValueError(
                f"Invalid annotation file. Task for video '{video_name}' is missing."
            )

        task: Dict[str, Any]
        for task in data:
            try:
                file = task["file_upload"]
            except KeyError as exc:
                raise ValueError(
                    "Invalid task. Key file_upload is missing for a task."
                ) from exc

            if file == video_name:
                return task

        raise ValueError(
            f"Invalid annotation file. Task for video '{video_name}' is missing."
        )

    def _parse_tracking_annotations_file(self) -> None:
        """
        Parse the tracking annotation JSON file.
        """

        task = self._find_task()

        annotations = task.get("annotations", [])

        if len(annotations) == 0:
            raise ValueError("Invalid task. Key annotations is missing or is empty.")

        for annotation in annotations:
            result = annotation.get("result", [])

            self._parse_tracking_annotations_result(result)

    def _parse_tracking_annotations_result(self, result: List[Dict[str, Any]]) -> None:
        """
        Parse a list of tracking annotation results.
        """

        if len(result) == 0:
            raise ValueError(
                "Invalid task. Key annotations.result is missing or is empty."
            )

        for res in result:
            value = res.get("value", {})

            frame_count = value.get("framesCount", -1)

            if frame_count != self.video_specs.frame_count:
                raise ValueError("Invalid task. Frame count differs.")

            labels = value.get("labels", [])

            if len(labels) == 0:
                raise ValueError(
                    "Invalid task. Key annotations.result.value.labels is missing or is empty."
                )

            boxes = value.get("sequence", [])

            if len(boxes) == 0:
                raise ValueError(
                    "Invalid task. Key annotations.result.value.sequence is missing or is empty."
                )

            for box in boxes:
                try:
                    frame_number = box["frame"]
                    x_min = box["x"]
                    y_min = box["y"]
                    width = box["width"]
                    height = box["height"]
                    rotation = box["rotation"]
                except KeyError as exc:
                    raise ValueError(
                        "Annotation must contain keys 'frame', 'time', 'x', 'y', 'width', "
                        "'height' and 'rotation'."
                    ) from exc

                self.annotations[frame_number].append(
                    LabelStudioTrackingTaskConverter.__Annotation(
                        x_min=x_min,
                        y_min=y_min,
                        width=width,
                        height=height,
                        rotation=rotation,
                        labels=labels,
                    )
                )

    def _extract_images(self) -> None:
        """
        Extract the images from the video that have at least one annotation.
        """

        video = cv2.VideoCapture(self.video_path)
        frame_counter = 0

        if video.isOpened():
            with TemporaryDirectory() as tmp_dir:
                with zipfile.ZipFile(self.image_archive_path, "w") as archive:
                    while True:
                        flag, frame = video.read()
                        if flag:
                            if frame_counter in self.annotations:
                                image_file = self._build_image_filename(
                                    frame_number=frame_counter
                                )
                                image_path = os.path.join(tmp_dir, image_file)
                                cv2.imwrite(image_path, frame)
                                archive.write(image_path, image_file)
                            frame_counter += 1
                        else:
                            break
        else:
            raise RuntimeError(f"Could not open video '{self.video_path}'")

    def _save_annotations(self) -> None:
        """
        Save the parsed annotations.
        """

        annotations = []
        for frame_number, annotation in self.annotations.items():
            results = [
                {
                    "original_width": self.video_specs.width,
                    "original_height": self.video_specs.height,
                    "image_rotation": 0,
                    "value": {
                        "x": a.x_min,
                        "y": a.y_min,
                        "width": a.width,
                        "height": a.height,
                        "rotation": a.rotation,
                        "rectanglelabels": a.labels,
                    },
                    "from_name": "label",
                    "to_name": "image",
                    "type": "rectanglelabels",
                    "origin": "manual",
                }
                for a in annotation
            ]

            annotation_dict = {
                "annotations": [
                    {
                        "result": results,
                    }
                ],
                "data": {"image": self._build_image_filename(frame_number)},
            }
            annotations.append(annotation_dict)

        with open(file=self.annotations_path, mode="w", encoding="utf-8") as file:
            json.dump(obj=annotations, fp=file, indent=2)

    def _build_image_filename(self, frame_number: int) -> str:
        """
        Build filename for an image based on video filename and frame number.
        """

        leading_zeros = len(str(self.video_specs.frame_count))
        stem = os.path.splitext(os.path.basename(self.video_path))[0]

        return f"{stem}_{frame_number:0{leading_zeros}d}.png"


def parse_args() -> argparse.Namespace:
    """
    Parse arguments for the CLI script.
    """

    parser = argparse.ArgumentParser(
        description="Tool to convert tracking task annotations."
    )

    parser.add_argument("video_path", help="The video file to extract keyframes from.")
    parser.add_argument(
        "tracking_annotations_path",
        help="The file containing the trackking annotation data.",
    )
    parser.add_argument(
        "image_archive_path", help="Zip file, the keyframes are written to."
    )
    parser.add_argument(
        "annotations_path",
        help="File the detection annotations are written to.",
    )

    return parser.parse_args(sys.argv[1:])


def main() -> None:
    """
    Main entrypoint for the CLI script.
    """

    args = parse_args()

    converter = LabelStudioTrackingTaskConverter(
        video_path=args.video_path,
        tracking_annotations_path=args.tracking_annotations_path,
        image_archive_path=args.image_archive_path,
        annotations_path=args.annotations_path,
    )

    converter.run()


if __name__ == "__main__":
    main()



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module for different utility operations regarding image (picture) objects"""

from __future__ import annotations

import copy
import logging
import os
import time
from typing import Any, Iterator, List, Optional, Tuple

import cv2
import imageio
import numpy as np
from mlcvzoo_base.utils.file_utils import ensure_dir

logger = logging.getLogger(__name__)


def _scale_frame(
    frame: np.ndarray, output_shape: int  # type: ignore[type-arg]
) -> Tuple[np.ndarray, Tuple[int, int]]:  # type: ignore[type-arg]
    """

    Args:
        frame: Numpy array, image
        output_shape: int, shape of the output

    Returns: Tuple of the scaled frame and shape of the frame in (x,y)

    """

    # TODO: check if this results in memory issues

    if output_shape > 0:
        scaled_frame = copy.deepcopy(frame)

        scale_factor = output_shape / scaled_frame.shape[0]
        resize_shape = (
            int(frame.shape[1] * scale_factor),
            int(frame.shape[0] * scale_factor),
        )

        scaled_frame = cv2.resize(
            scaled_frame, resize_shape, interpolation=cv2.INTER_AREA
        )
        return scaled_frame, resize_shape
    else:
        return frame, (frame.shape[1], frame.shape[0])


class VideoLiveOutput:
    """Class for displaying a video"""

    MODE_TERMINATE = "terminate"
    MODE_GO = "go"
    MODE_STEP = "step"

    KEY_CLOSE = "q"
    KEY_STEP = "s"
    KEY_GO = "c"
    KEY_NEXT = "n"

    MODES = [MODE_TERMINATE, MODE_GO, MODE_STEP]

    # TODO: Better name for use_mode?
    def __init__(
        self,
        use_mode: bool = True,
        show_fps: bool = False,
        window_name: str = "Video stream",
        resize_window: bool = True,
        window_size: int = 500,
        mode: Optional[str] = None,
    ):
        self.use_mode = use_mode
        self.show_fps = show_fps
        self.window_name = window_name
        self.resize_window = resize_window
        self.window_size = window_size

        self.mode = VideoLiveOutput.MODE_STEP
        if mode is not None:
            if mode not in self.MODES:
                raise ValueError(
                    "Invalid mode='%s' has to be one of '%s'" % (mode, self.MODES)
                )

            self.mode = mode
        self.frame = 0
        self.resized_window = False
        self.current_fps = 0.0

        self.start = time.time()
        self.end = time.time() - self.start

        cv2.namedWindow(self.window_name, cv2.WINDOW_NORMAL)

        logger.info(
            "Create VideoLiveOutput with title '%s'. "
            "Step and continuous mode.\n"
            "Press %s to continue, %s to step. "
            "Press %s for next frame.",
            self.window_name,
            VideoLiveOutput.KEY_GO,
            VideoLiveOutput.KEY_STEP,
            VideoLiveOutput.KEY_NEXT,
        )

    def output_frame(self, frame: np.ndarray, img_info: str = "") -> None:  # type: ignore[type-arg]
        """
        Creates and shows the output frame based on class attributes and parameters

        Args:
            frame: Numpy array, the frame
            img_info: String, any information that should be printed on the image

        Returns: None

        """

        if self.mode != VideoLiveOutput.MODE_TERMINATE:
            self.end = time.time() - self.start

            # will only be executed once
            if not self.resized_window:
                scale_factor = frame.shape[1] / frame.shape[0]

                if self.resize_window is True:
                    window_height = self.window_size
                else:
                    window_height = frame.shape[0]

                cv2.resizeWindow(
                    winname=self.window_name,
                    height=window_height,
                    width=int(window_height * scale_factor),
                )

                self.resized_window = True

            if self.show_fps:
                self.current_fps = 1 / self.end if self.end != 0.0 else 0.0

                # put image name in visible image
                cv2.putText(
                    img=frame,
                    text="{:.4f}".format(self.current_fps),
                    org=(20, 20),
                    fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                    fontScale=0.75,
                    color=(255, 255, 255),
                    thickness=2,
                    lineType=cv2.LINE_AA,
                )

            if img_info != "":
                # put image name in visible image
                cv2.putText(
                    img=frame,
                    text="{}".format(os.path.basename(img_info)),
                    org=(20, 40),
                    fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                    fontScale=0.75,
                    color=(255, 255, 255),
                    thickness=2,
                    lineType=cv2.LINE_AA,
                )

            cv2.imshow(self.window_name, frame)
            key = 0

            if self.use_mode:
                while self.mode == VideoLiveOutput.MODE_STEP and key != ord(
                    VideoLiveOutput.KEY_NEXT
                ):
                    key = cv2.waitKey(0) & 0xFF
                    if key == ord(VideoLiveOutput.KEY_GO):
                        self.mode = VideoLiveOutput.MODE_GO

                    if key == 27 or key == ord(VideoLiveOutput.KEY_CLOSE):
                        self.close()

                if self.mode == VideoLiveOutput.MODE_GO:
                    key = cv2.waitKey(1) & 0xFF

                if key == ord(VideoLiveOutput.KEY_STEP):
                    self.mode = VideoLiveOutput.MODE_STEP

                if key == 27 or key == ord(VideoLiveOutput.KEY_CLOSE):
                    self.close()

            self.frame += 1

            self.start = time.time()

    def close(self) -> None:
        """Closes open windows"""

        self.mode = VideoLiveOutput.MODE_TERMINATE
        cv2.destroyAllWindows()

    def is_terminated(self) -> bool:
        """Returns current termination status of the class"""

        return self.mode == VideoLiveOutput.MODE_TERMINATE


class VideoFileInput:
    """Class for handling a video and navigating through its frames"""

    def __init__(self, path: str):
        logger.debug("Init VideoFileInput for video '%s'", path)

        self.cap = cv2.VideoCapture(path)

        self.current_frame: Optional[np.ndarray] = None  # type: ignore[type-arg]

    def goto_frame(self, num: int) -> None:
        """Sets current frame to given position

        Args:
            num: int, number of frame that is to be reached

        Returns: None

        """

        self.cap.set(cv2.CAP_PROP_POS_FRAMES, num)

    def next_frame(self) -> Iterator[Optional[np.ndarray]]:  # type: ignore[type-arg]
        """
        Jumps to the next frame

        Returns: Iterator object based on the current frame

        """

        while True:
            finished_stream, self.current_frame = self.get_image()
            if not finished_stream:
                self.current_frame = None
                break

            yield self.current_frame
        yield self.current_frame

    def get_image(self, frame_num: Optional[int] = None) -> Any:
        """
        Returns image at the given frame or the next video frame

        Args:
            frame_num: Optional int, position of frame

        Returns: Any, either the requested frame or the next video frame

        """

        if frame_num is not None:
            self.goto_frame(frame_num)

        return self.cap.read()

    def has_next(self) -> bool:
        """
        Checks whether the current frame is available

        Returns: Bool, the availability of current_frame attribute
        """

        return self.current_frame is not None

    def skip_frames(self, number_frames: int) -> None:
        """
        Skips the given amount of frames, sets the current_frame attribute to the new frame

        Args:
            number_frames: int, the number of frames to be skipped

        Returns: None

        """

        for i in range(0, number_frames):
            finished_stream = self.cap.grab()

            if finished_stream:
                self.current_frame = None

    def number_of_frames(self) -> int:
        """Returns: int, number of overall frames in the video"""

        return int(self.cap.get(cv2.CAP_PROP_FRAME_COUNT))

    def __del__(self):  # type: ignore
        self.cap.release()


class GifFileOutput:
    """Class for creating and saving gifs"""

    gif_frame_list: List[np.ndarray] = []  # type: ignore[type-arg]

    def __init__(
        self,
        output_path: str,
        fps: int,
        output_shape: int,
        ask_for_dir_creation: bool = True,
    ):
        self.output_path = output_path
        self.fps = fps
        self.output_shape = output_shape
        self.ask_for_dir_creation = ask_for_dir_creation

    def add_frame(self, frame: np.ndarray) -> None:  # type: ignore[type-arg]
        """
        Adds a frame to the gif_frame_list attribute
        Args:
            frame: Numpy array, the frame to be added

        Returns: None

        """

        scaled_frame, _ = _scale_frame(frame=frame, output_shape=self.output_shape)

        self.gif_frame_list.append(scaled_frame)

    def save_gif(self) -> None:
        """
        Saves the gif_frame_list to a file specified in the output_path attribute

        Returns: None

        """

        if len(self.gif_frame_list) > 0:
            self.__ensure_gif_dir()

            if os.path.isdir(os.path.dirname(self.output_path)):
                logger.debug("Creating gif ...")
                imageio.mimsave(
                    uri=self.output_path, ims=self.gif_frame_list, fps=self.fps  # type: ignore
                )
                logger.info("Saved gif to: %s", self.output_path)
        else:
            logger.warning(
                "WARNING: image list for gif creation is emtpy! Will not generate gif..."
            )

    def __ensure_gif_dir(self) -> None:
        if not os.path.isdir(os.path.dirname(self.output_path)):
            if self.ask_for_dir_creation:
                logger.info(
                    "Create directory for gif-path: %s?\n <y> or <n>",
                    self.output_path,
                )

                keyboard_input = input()

                if keyboard_input == "y":
                    os.makedirs(name=self.output_path, exist_ok=True)


class VideoFileOutput:
    """Class for writing a video to a file"""

    def __init__(
        self,
        path: str,
        output_shape: int,
        fps: int = 25,
    ):
        self.path = path
        self.fps = fps
        self.output_shape = output_shape

        self.open = False

        assert len(path) > 0

    def add_frame(self, frame: np.ndarray, img_info: str = "") -> None:  # type: ignore[type-arg]
        """
        Adds a frame to a video by adding frame information if available, resizing the frame
        and writing it to the location specified in path attribute

        Args:
            frame: Numpy array, the frame to be added
            img_info: String, any information that should be printed on the image

        Returns: None

        """

        if not self.open:
            ensure_dir(file_path=self.path, verbose=True)

            logger.info("Open VideoWriter for video-path: %s", self.path)

            scaled_frame, resize_shape = _scale_frame(
                frame=frame, output_shape=self.output_shape
            )

            self.video_writer = cv2.VideoWriter(
                filename=self.path,
                fourcc=cv2.VideoWriter_fourcc(*"mp4v"),  # type: ignore[attr-defined]
                fps=self.fps,
                frameSize=(
                    resize_shape[0],
                    resize_shape[1],
                ),  # opencv uses (y,x) format
            )
            # filename, fourcc, fps, frameSize, isColor = None
            self.open = True

        if img_info != "":
            # put image name in visible image
            cv2.putText(
                img=frame,
                text="{}".format(os.path.basename(img_info)),
                org=(20, 40),
                fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                fontScale=0.75,
                color=(255, 255, 255),
                thickness=2,
                lineType=cv2.LINE_AA,
            )

        scaled_frame, resize_shape = _scale_frame(
            frame=frame, output_shape=self.output_shape
        )
        self.video_writer.write(scaled_frame)

    def close(self) -> None:
        """Closing the VideoWriter"""

        if self.open:
            logger.info("Write video to: %s", self.path)
            self.video_writer.release()



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Module for configuring the parsing of information from yaml in python
accessible attributes for the Evaluation Runner (ER) class
"""

from typing import List, Optional

import related
from attr import define
from config_builder import BaseConfigClass
from mlcvzoo_base.configuration.annotation_handler_config import AnnotationHandlerConfig
from mlcvzoo_base.configuration.mlfow_config import MLFlowConfig
from mlcvzoo_base.configuration.model_config import ModelConfig
from mlcvzoo_base.evaluation.object_detection.configuration import (
    TensorboardLoggingConfig,
)

from mlcvzoo_util.model_evaluator.structs import CheckpointLoggingModes


@define
class ModelEvaluatorMLflowConfig(BaseConfigClass):
    """Class for parsing information for mlflow"""

    __related_strict__ = True

    config: MLFlowConfig = related.ChildField(cls=MLFlowConfig)

    checkpoint_log_mode: str = related.StringField(
        required=False, default=CheckpointLoggingModes.NONE.value
    )


@define
class CheckpointConfig(BaseConfigClass):
    """Class for parsing information about model checkpoints to evaluate"""

    __related_strict__ = True

    checkpoint_dir: str = related.StringField()
    checkpoint_filename_suffix: str = related.StringField()
    ignore: List[str] = related.SequenceField(cls=str, required=False, default=[])


@define
class ModelEvaluatorConfig(BaseConfigClass):
    """Class for parsing information from yaml in respective hierarchy"""

    __related_strict__ = True

    iou_thresholds: List[float] = related.SequenceField(cls=float)

    checkpoint_config: Optional[CheckpointConfig] = related.ChildField(
        cls=CheckpointConfig, required=False, default=None
    )

    mlflow_config: Optional[ModelEvaluatorMLflowConfig] = related.ChildField(
        cls=ModelEvaluatorMLflowConfig, required=False, default=None
    )

    tensorboard_logging_config: Optional[TensorboardLoggingConfig] = related.ChildField(
        cls=TensorboardLoggingConfig, required=False, default=None
    )


@define
class ModelEvaluatorCLIConfig(ModelEvaluatorConfig):
    """Class for parsing information from yaml in respective hierarchy"""

    __related_strict__ = True

    iou_thresholds: List[float] = related.SequenceField(float)

    model_config: ModelConfig = related.ChildField(cls=ModelConfig)

    checkpoint_config: CheckpointConfig = related.ChildField(cls=CheckpointConfig)

    annotation_handler_config: AnnotationHandlerConfig = related.ChildField(
        cls=AnnotationHandlerConfig
    )

    mlflow_config: Optional[ModelEvaluatorMLflowConfig] = related.ChildField(
        cls=ModelEvaluatorMLflowConfig, required=False, default=None
    )

    tensorboard_logging_config: Optional[TensorboardLoggingConfig] = related.ChildField(
        cls=TensorboardLoggingConfig, required=False, default=None
    )



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Module for defining all configuration classes that
are used in the context of the CVATAnnotationHandler
"""

import logging
import os
import sys
from getpass import getpass
from typing import List, Optional

import related
from config_builder import BaseConfigClass

logger = logging.getLogger(__name__)


@related.mutable(strict=True)
class CVATCLIConfig(BaseConfigClass):
    """
    Configuration class for defining parameters that are needed for
    the CVAT commandline-interface
    """

    cli_path: str = related.StringField()
    server_host: str = related.StringField()
    auth: str = related.StringField()
    disable_ssl_verify: bool = related.BooleanField(required=False, default=False)
    server_port: Optional[int] = related.ChildField(
        cls=int, required=False, default=None
    )
    password_path: Optional[str] = related.ChildField(
        cls=str, required=False, default=None
    )
    python_executable: str = related.StringField(required=False, default=sys.executable)

    def create_base_command_string(self) -> str:
        """
        Create the base string that is needed to call the CVAT commandline-interface

        Returns:
            The created string
        """
        auth: str = self.auth

        if ":" not in auth:
            if self.password_path is not None:
                if os.path.isfile(self.password_path):
                    with open(self.password_path, "r") as password_file:
                        password = password_file.readline().strip()

                        auth = f"{self.auth}:{password}"
                else:
                    if self.password_path != "":
                        logger.warning(
                            "The provided password-path does not exists! "
                            "Therefore, only the provided cvat_cli_config.auth config attribute "
                            "will be used! password_path: '%s'" % self.password_path
                        )
            else:
                logger.info(
                    "Neither a password is given in the auth parameter, nor a "
                    "password file is given, please enter the password"
                )
                password = getpass()
                auth = f"{auth}:{password}"

        if self.server_port is not None:
            base_command = f"%s %s --server-host %s --server-port %s --auth %s" % (
                self.python_executable,
                self.cli_path,
                self.server_host,
                self.server_port,
                auth,
            )
        else:
            base_command = "%s %s --server-host %s --auth %s" % (
                self.python_executable,
                self.cli_path,
                self.server_host,
                auth,
            )

        return base_command


@related.mutable(strict=True)
class CVATTaskInfoConfig(BaseConfigClass):
    """
    Configuration class that defines attributes that are specific for
    a CVAT task
    """

    VALID_FORMATS = [
        "COCO 1.0",
        "CVAT for images 1.1",
        "CVAT for video 1.1",
        "Datumaro 1.0",
        "LabelMe 3.0",
        "MOT 1.1",
        "MOTS PNG 1.0",
        "PASCAL VOC 1.1",
        "Segmentation mask 1.1",
        "TFRecord 1.0",
        "YOLO 1.1",
        "ImageNet 1.0",
        "CamVid 1.0",
    ]

    task_ID: int = related.IntegerField()
    annotation_format: str = related.StringField(
        required=False, default="PASCAL VOC 1.1"
    )

    def check_values(self) -> bool:
        return self.annotation_format in self.VALID_FORMATS


@related.mutable(strict=True)
class CVATTaskDumpConfig(BaseConfigClass):
    """
    Configuration class that defines attributes that are specific for a
    dump of a CVAT task.
    """

    task_info: CVATTaskInfoConfig = related.ChildField(cls=CVATTaskInfoConfig)

    # if value is empty, the task will be downloaded to a tmp-file
    target_zip_path: str = related.StringField(required=False, default="")

    # Whether to extract or not the data of the downloaded zip file.
    # The data is extracted to the directory with the base-name of the "target_zip_path"
    extract_data: bool = related.BooleanField(required=False, default=True)

    zip_extract_dir: str = related.StringField(required=False, default="")

    move_extracted_data_to_dataset_dir: bool = related.BooleanField(
        required=False, default=False
    )

    # directory where the xml-files of the downloaded zip-file should be copied to,
    # in order to generate an overall directory for tasks which share the same directory-tree
    # NOTE: this will trigger the extraction of the downloaded zip-file
    dataset_dir: str = related.StringField(required=False, default="")

    # whether to use trigger an new download of a zip-file or to use an already
    # existing zip-file (when it exists)
    overwrite_existing_zip: bool = related.BooleanField(required=False, default=False)

    # delete all generated files during the dumping-routine of this CVAT task
    clean_up_dump_data: bool = related.BooleanField(required=False, default=False)


@related.mutable(strict=True)
class CVATTaskUploadConfig(BaseConfigClass):
    """
    Configuration class that defines attributes that are specific for an
    upload of a CVAT task.
    """

    task_info: CVATTaskInfoConfig = related.ChildField(cls=CVATTaskInfoConfig)

    execute_upload: bool = related.BooleanField()

    # is required because the information of "labelmap.txt" and "Main/default.txt" has to
    # be known for creating a zip-file which can be uploaded.
    source_zip_path: str = related.StringField()

    # whether to use the annotations from the dumped zip-file
    # or to use the annotations provided in the "prediction_data_dir"
    use_prediction_data: bool = related.BooleanField()

    # directory where generated annotations are stored
    prediction_data_dir: str = related.StringField(required=False, default="")

    # by default the "source_zip_path" is used with an "_upload" extension
    # This path defines the location for the zip-file which will be uploaded to the CVAT task
    target_zip_path: str = related.StringField(required=False, default="")

    # delete all generated files during the dumping-routine of this CVAT task
    clean_up_dump_data: bool = related.BooleanField(required=False, default=False)

    # delete all generated files during the uploaded-routine of this CVAT task
    clean_up_upload_data: bool = related.BooleanField(required=False, default=False)


@related.mutable(strict=False)
class CVATAnnotationHandlerConfig(BaseConfigClass):
    """
    Central configuration class for the CVATAnnotationHandler
    """

    cvat_cli_config: CVATCLIConfig = related.ChildField(cls=CVATCLIConfig)

    execute_tasks_separately: bool = related.BooleanField(required=False, default=True)

    dump_task_configs: List[CVATTaskDumpConfig] = related.SequenceField(
        cls=CVATTaskDumpConfig, required=False, default=[]
    )
    upload_task_configs: List[CVATTaskUploadConfig] = related.SequenceField(
        cls=CVATTaskUploadConfig, required=False, default=[]
    )



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Module for defining enums and structs that are used across the
model_evaluator package.
"""
from dataclasses import dataclass
from enum import Enum, auto


class CheckpointLoggingModes(Enum):
    BEST = "best"
    ALL = "all"
    NONE = "none"


@dataclass
class CheckpointInfo:
    """
    Dataclass for storing information about a checkpoint
    and its overall AP metric
    """

    path: str
    score: float
